# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information


import os
import sys
from datetime import date
sys.path.insert(0, os.path.abspath("../src"))
import curses_fzf

first_release_year = "2026"
now_year = str(date.today().year)
copyright_year = first_release_year
if first_release_year != now_year:
    copyright_year = f"{first_release_year}-{now_year}"

project = 'curses-fzf'
author = 'Heiko Finzel'
copyright = f'{copyright_year}, {author}'
release = curses_fzf.__version__

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    'myst_parser',  # mardown support
    'sphinx.ext.autodoc',  # read docstrings from code
    'sphinx.ext.autodoc.typehints',  # show type hints in the description of the function, not in the signature
    'sphinx.ext.napoleon',  # support for Google and NumPy style docstrings
    'sphinx.ext.viewcode',  # add links to highlighted source code of documented Python objects
    'sphinx.ext.inheritance_diagram',  # graphviz diagrams for class inheritance
    'sphinx.ext.graphviz',  # graphviz diagrams for class inheritance
]

autodoc_preserve_defaults = True

templates_path = ['_templates']
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store']


# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = 'sphinx_rtd_theme'
html_static_path = ['_static']



import curses
from unittest.mock import MagicMock, patch, call
from curses_fzf import ColorTheme
from curses_fzf.help import _help, _base_window


def test_base_window_renders_correctly():
    mock_stdscr = MagicMock(spec=curses.window)
    mock_stdscr.getmaxyx.return_value = (15, 25)

    with patch('curses.color_pair') as mock_color_pair, \
            patch('curses.textpad') as mock_textpad:

        mock_color_pair.side_effect = lambda x: x   # return color pair ID directly

        result = _base_window(
            mock_stdscr,
            "MY TITLE",
            "My Footer is also very long and should be truncated",
            ColorTheme()
        )
        assert result == (15, 25)
        mock_stdscr.clear.assert_called_once()
        mock_stdscr.getmaxyx.assert_called_once()
        mock_textpad.rectangle.assert_called_once_with(mock_stdscr, 1, 0, 15-2, 25-1)
        mock_stdscr.addstr.assert_has_calls([
            call(1, 2, " MY TITLE ", 33),
            call(14, 2, "My Footer is also ver", 33),
        ], any_order=True)


def test_help():
    mock_stdscr = MagicMock(spec=curses.window)
    mock_stdscr.getmaxyx.return_value = (21, 40)
    mock_stdscr.getch.return_value = curses.KEY_F1  # simulate pressing F1 to exit help

    with patch('curses.color_pair') as mock_color_pair, \
            patch('curses.textpad') as mock_textpad:

        mock_color_pair.side_effect = lambda x: x   # return color pair ID directly

        _help(mock_stdscr, page_size=5, color_theme=ColorTheme())

        mock_stdscr.clear.assert_called_once()
        mock_stdscr.getmaxyx.assert_called_once()
        mock_textpad.rectangle.assert_called_once_with(mock_stdscr, 1, 0, 21-2, 40-1)
        mock_stdscr.addstr.assert_has_calls([
            call(1, 2, " HELP ", 33),
            call(20, 2, "F1 = close help", 33),
            call(5, 7, "text characters", 37),
        ], any_order=True)



import pytest
from curses_fzf import ScoringResult, scoring_full_words


@pytest.fixture
def fox():
    return ScoringResult("fox bro", "The quick brown fox jumps over the lazy dog and lands elegantly on the meadow.")


@pytest.fixture
def banana():
    return ScoringResult("app", "BaNaNa")


@pytest.fixture
def henry():
    return ScoringResult("me", "Henry is at home and watches a snowman melt slowly.")


def test_scoringresult_init(fox, banana):
    assert fox.query == "fox bro"
    assert fox.candidate == "The quick brown fox jumps over the lazy dog and lands elegantly on the meadow."
    assert banana.query == "app"
    assert banana.candidate == "BaNaNa"
    assert fox.query_lower == "fox bro"
    assert fox.candidate_lower == "the quick brown fox jumps over the lazy dog and lands elegantly on the meadow."
    assert banana.query_lower == "app"
    assert banana.candidate_lower == "banana"
    assert fox.query_words_with_index == [('fox', 0), ('bro', 4)]
    assert banana.query_words_with_index == [("app", 0)]
    assert fox.candidate_words_with_index == [('the', 0), ('quick', 4), ('brown', 10), ('fox', 16), ('jumps', 20),
                                              ('over', 26), ('the', 31), ('lazy', 35), ('dog', 40), ('and', 44),
                                              ('lands', 48), ('elegantly', 54), ('on', 64), ('the', 67),
                                              ('meadow.', 71)]
    assert banana.candidate_words_with_index == [('banana', 0)]
    assert fox.score == 0
    assert banana.score == 0
    assert fox.matches == []
    assert banana.matches == []
    assert fox._already_matched_words == set()
    assert banana._already_matched_words == set()


def test_scoringresult_int(fox, banana):
    assert int(fox) == 0
    assert int(banana) == 0
    fox.score = 42
    banana.score = 7
    assert int(fox) == 42
    assert int(banana) == 7


def test_scoringresult_lt(fox, banana):
    fox.score = 10
    banana.score = 20
    assert (fox < banana) == (fox.score < banana.score)
    assert (banana < fox) == (banana.score < fox.score)
    assert (fox < banana) == True  # noqa: E712
    assert (banana < fox) == False  # noqa: E712
    assert (fox < 42) == True  # noqa: E712
    assert (banana < 15) == False  # noqa: E712
    with pytest.raises(TypeError):
        fox < "foo"


def test_scoringresult_gt(fox, banana):
    fox.score = 10
    banana.score = 20
    assert (fox > banana) == (fox.score > banana.score)
    assert (banana > fox) == (banana.score > fox.score)
    assert (fox > banana) == False  # noqa: E712
    assert (banana > fox) == True  # noqa: E712
    assert (fox > 42) == False  # noqa: E712
    assert (banana > 15) == True  # noqa: E712
    with pytest.raises(TypeError):
        fox > "foo"


def test_scoringresult_eq(fox, banana):
    fox.score = 10
    banana.score = 10
    assert (fox == banana) == (fox.score == banana.score)
    assert (banana == fox) == (banana.score == fox.score)
    assert (fox == banana) == True  # noqa: E712
    assert (banana == fox) == True  # noqa: E712
    assert (fox == 42) == False  # noqa: E712
    assert (banana == 10) == True  # noqa: E712
    banana.score = 20
    assert (fox == banana) == False  # noqa: E712
    assert (banana == fox) == False  # noqa: E712
    assert (fox == "foo") == False  # noqa: E712


def test_scoringresult_add_match(fox, banana):
    fox.add_match(16, 3, 10)
    assert fox.score == 10
    assert fox.matches == [(16, 3)]
    fox.add_match(27, 5, 15)
    assert fox.score == 25
    assert fox.matches == [(16, 3), (27, 5)]
    banana.add_match(0, 6, 5)
    assert banana.score == 5
    assert banana.matches == [(0, 6)]
    banana.add_match(55, 2, 12)
    assert banana.score == 17
    assert banana.matches == [(0, 6), (55, 2)]


def test_scoringresult_find_best_word_match(fox, henry):
    assert fox.find_best_word_match("bro") == ('brown', 10, 0, 60)
    # since "brown" already matched for "bro"
    assert fox.find_best_word_match("own") is None
    assert fox.find_best_word_match("fox") == ('fox', 16, 0, 100)
    assert fox.find_best_word_match("tly") == ('elegantly', 54, 6, 33)
    assert fox.find_best_word_match("cat") is None
    # melt > home because the match is closer to the beginning of the word
    assert henry.find_best_word_match("me") == ('melt', 39, 0, 50)


def test_scoring_full_words(henry, fox, banana):
    result = scoring_full_words(henry.query, henry.candidate)
    # 50 (% matched word "melt") * 1.5 (matched word is the beginning of the candidate word)
    # * 1.2 (matched words are in the original order) = 90
    assert result.score == 90
    assert result.matches == [(39, henry.query)]
    assert result._already_matched_words == {39}
    result = scoring_full_words("tch is he", henry.candidate)
    assert result.score == 84
    assert result.matches == [(23, "tch"), (6, "is"), (0, "he")]
    assert result._already_matched_words == {0, 6, 21}
    result = scoring_full_words(fox.query, fox.candidate)
    assert result.score == 120
    assert result.matches == [(16, "fox"), (10, "bro")]
    assert result._already_matched_words == {10, 16}
    # always the same result if the query is empty
    result = scoring_full_words("", fox.candidate)
    assert result.score == 100
    assert result.matches == []
    assert result._already_matched_words == set()
    result = scoring_full_words("", "foo")
    assert result.score == 100
    result = scoring_full_words("", "bar baz")
    assert result.score == 100
    # no match
    result = scoring_full_words(banana.query, banana.candidate)
    assert result.score == 0
    assert result.matches == []
    assert result._already_matched_words == set()



from .__about__ import __version__
from .colors import Color, ColorTheme
from .errors import CursesFzfException, CursesFzfAborted, CursesFzfAssertion, CursesFzfIndexOutOfBounds
from .fuzzyfinder import FuzzyFinder
from .scoring import ScoringResult, scoring_full_words

__all__ = [
    "__version__",
    "Color",
    "ColorTheme",
    "CursesFzfException",
    "CursesFzfAborted",
    "CursesFzfAssertion",
    "CursesFzfIndexOutOfBounds",
    "FuzzyFinder",
    "ScoringResult",
    "scoring_full_words",
]



import curses
import curses.textpad
from typing import Tuple

from .colors import ColorTheme


def _base_window(stdscr: curses.window, title: str, footer: str, color_theme: ColorTheme) -> Tuple[int, int]:
    """
    Draw a basic window with frame, header and footer line.
    """
    stdscr.clear()
    height, width = stdscr.getmaxyx()
    curses.textpad.rectangle(stdscr, 1, 0, height-2, width-1)
    stdscr.addstr(1, 2, " " + title[:width-6] + " ", curses.color_pair(color_theme.window_title))
    stdscr.addstr(height-1, 2, footer[:width-4], curses.color_pair(color_theme.footer))
    return height, width


def _help(stdscr: curses.window, page_size: int, color_theme: ColorTheme) -> None:
    """
    Print a help screen.
    """
    help = (
        ("Fuzzy Finder Query", (
            ("text characters", "Enter a fuzzy finder query."),
            ("BACKSPACE", "Remove one character before the cursor."),
            ("DELETE", "Remove one character at the cursor."),
            ("CTRL + K", "Clear entire query."),
            ("ARROW-LEFT", "Move cursor left 1 position in query."),
            ("ARROW-RIGHT", "Move cursor right 1 position in query."),
        )),
        ("List Movement", (
            ("ARROW-UP", "Move up 1 entry."),
            ("ARROW-DOWN", "Move down 1 entry."),
            ("PAGE-UP", f"Move up {page_size} entries."),
            ("PAGE-DOWN", f"Move down {page_size} entries."),
            ("HOME", "Move to first item."),
            ("END", "Move to last item."),
        )),
        ("Item Selection (multi-select only)", (
            ("TAB", "Toggle selection of the current item."),
            ("CTRL + A", "Select all items matching current filter query."),
            ("CTRL + X", "Deselect all items matching current filter query."),
        )),
        ("Control Commands", (
            ("ENTER", "Accept the current item(s)."),
            ("ESC", "Abort fuzzy raising CursesFzfAborted exception."),
            ("CTRL + P", "Toggle preview window (if a preview function is provided)."),
            ("F1", "Toggle this help screen."),
        )),
    )
    while True:
        height, width = _base_window(stdscr, "HELP", "F1 = close help", color_theme)
        line = 2
        section_start = 5
        col1_start = section_start + 2
        col1_width = 15
        for section in help:
            if line > height-4:
                break
            line += 1
            # print section, limit width to one space before frame
            stdscr.addstr(line, section_start, str(section[0][:width - section_start - 2]),
                          curses.color_pair(color_theme.text))
            line += 2
            for key in section[1]:
                if line > height-4:
                    break
                # print key-column with wixed width
                stdscr.addstr(line, col1_start, str(key[0][:min(col1_width, width - col1_start - 2)]),
                              curses.color_pair(color_theme.text))
                # print key description, limit width to one space before frame
                if width > col1_width + col1_start + 3:
                    stdscr.addstr(line, col1_start + col1_width + 2, str(key[1][:width - col1_start - col1_width - 4]),
                                  curses.color_pair(color_theme.text))
                line += 1
        stdscr.refresh()
        if stdscr.getch() == curses.KEY_F1:
            return



import pytest
import curses
from unittest.mock import MagicMock, patch, call
from curses_fzf import FuzzyFinder, ScoringResult, CursesFzfAborted, CursesFzfAssertion, CursesFzfIndexOutOfBounds


def test_kb_move_items_cursor_absolute():
    sr = ScoringResult("", "")
    fzf = FuzzyFinder()
    fzf.filtered = [("item1", sr), ("item2", sr), ("item3", sr)]
    assert fzf.cursor_items == 0
    fzf.kb_move_items_cursor_absolute(1)
    assert fzf.cursor_items == 1
    fzf.kb_move_items_cursor_absolute(2)
    assert fzf.cursor_items == 2
    fzf.kb_move_items_cursor_absolute(-1)
    assert fzf.cursor_items == 0
    fzf.kb_move_items_cursor_absolute(99)
    assert fzf.cursor_items == 2
    fzf.kb_move_items_cursor_absolute(0)
    assert fzf.cursor_items == 0


def test_kb_move_cursor_relative():
    sr = ScoringResult("", "")
    fzf = FuzzyFinder()
    fzf.filtered = [("item1", sr), ("item2", sr), ("item3", sr)]
    assert fzf.cursor_items == 0
    fzf.kb_move_items_cursor_relative(1)
    assert fzf.cursor_items == 1
    fzf.kb_move_items_cursor_relative(1)
    assert fzf.cursor_items == 2
    fzf.kb_move_items_cursor_relative(1)
    assert fzf.cursor_items == 2
    fzf.kb_move_items_cursor_relative(-1)
    assert fzf.cursor_items == 1
    fzf.kb_move_items_cursor_relative(-1)
    assert fzf.cursor_items == 0
    fzf.kb_move_items_cursor_relative(-1)
    assert fzf.cursor_items == 0
    fzf.kb_move_items_cursor_relative(10)
    assert fzf.cursor_items == 2
    fzf.kb_move_items_cursor_relative(-10)
    assert fzf.cursor_items == 0


def test_kb_move_query_cursor_absolute():
    fzf = FuzzyFinder(query="my initial query")
    assert fzf.cursor_query == 16
    fzf.kb_move_query_cursor_absolute(0)
    assert fzf.cursor_query == 0
    fzf.kb_move_query_cursor_absolute(-1)
    assert fzf.cursor_query == 0
    fzf.kb_move_query_cursor_absolute(99)
    assert fzf.cursor_query == 16
    fzf.kb_move_query_cursor_absolute(2)
    assert fzf.cursor_query == 2


def test_kb_move_query_cursor_relative():
    fzf = FuzzyFinder(query="my initial query")
    assert fzf.cursor_query == 16
    fzf.kb_move_query_cursor_relative(1)
    assert fzf.cursor_query == 16
    fzf.kb_move_query_cursor_relative(-1)
    assert fzf.cursor_query == 15
    fzf.kb_move_query_cursor_relative(-10)
    assert fzf.cursor_query == 5
    fzf.kb_move_query_cursor_relative(-10)
    assert fzf.cursor_query == 0
    fzf.kb_move_query_cursor_relative(10)
    assert fzf.cursor_query == 10
    fzf.kb_move_query_cursor_relative(1)
    assert fzf.cursor_query == 11
    fzf.kb_move_query_cursor_relative(10)
    assert fzf.cursor_query == 16


def test_kb_abort_selection():
    fzf = FuzzyFinder()
    with pytest.raises(CursesFzfAborted):
        fzf.kb_abort_selection()


def test_kb_accept_selection():
    sr = ScoringResult("", "")
    fzf_single = FuzzyFinder()
    assert fzf_single.return_selection_now == False  # noqa: E712
    # don't allow accept in single mode if no item in filtered list
    fzf_single.filtered = []
    fzf_single.kb_accept_selection()
    assert fzf_single.return_selection_now == False  # noqa: E712
    # accept selection
    fzf_single.filtered = [("item1", sr), ("item2", sr), ("item3", sr)]
    fzf_single.kb_accept_selection()
    assert fzf_single.return_selection_now == True  # noqa: E712
    fzf_multi = FuzzyFinder(multi=True)
    assert fzf_multi.return_selection_now == False  # noqa: E712
    # allow accept in multi mode even if no item in filtered list
    fzf_multi.filtered = []
    fzf_multi.kb_accept_selection()
    assert fzf_multi.return_selection_now == True  # noqa: E712


def test_kb_toggle_preview():
    fzf = FuzzyFinder()
    assert fzf.show_preview == True  # noqa: E712
    fzf.kb_toggle_preview()
    assert fzf.show_preview == False  # noqa: E712
    fzf.kb_toggle_preview()
    assert fzf.show_preview == True  # noqa: E712


def test_kb_toggle_selection():
    sr = ScoringResult("", "")
    fzf = FuzzyFinder(multi=True)
    fzf.filtered = [("item1", sr), ("item2", sr), ("item3", sr)]
    assert fzf.selected == []
    fzf.kb_toggle_selection()
    assert fzf.selected == ["item1"]
    fzf.kb_toggle_selection()
    assert fzf.selected == []
    fzf.kb_move_items_cursor_relative(1)
    fzf.kb_toggle_selection()
    assert fzf.selected == ["item2"]
    fzf.kb_move_items_cursor_relative(1)
    fzf.kb_toggle_selection()
    assert fzf.selected == ["item2", "item3"]
    fzf.kb_move_items_cursor_relative(-1)
    fzf.kb_toggle_selection()
    assert fzf.selected == ["item3"]
    fzf.kb_move_items_cursor_relative(-1)
    fzf.kb_toggle_selection()
    assert fzf.selected == ["item3", "item1"]
    fzf.kb_move_items_cursor_absolute(99)
    fzf.kb_toggle_selection()
    assert fzf.selected == ["item1"]
    fzf.kb_move_items_cursor_absolute(0)
    fzf.kb_toggle_selection()
    assert fzf.selected == []


def test_kb_select_all():
    sr = ScoringResult("", "")
    fzf = FuzzyFinder(multi=True)
    fzf.filtered = [("item1", sr), ("item2", sr), ("item3", sr)]
    assert fzf.selected == []
    fzf.kb_select_all()
    assert fzf.selected == ["item1", "item2", "item3"]
    fzf.kb_move_items_cursor_relative(1)
    fzf.kb_select_all()
    assert fzf.selected == ["item1", "item2", "item3"]
    fzf.selected = ["item4", "item5"]
    fzf.kb_select_all()
    assert fzf.selected == ["item4", "item5", "item1", "item2", "item3"]


def test_kb_deselect_all():
    sr = ScoringResult("", "")
    fzf = FuzzyFinder(multi=True)
    fzf.filtered = [("item1", sr), ("item2", sr), ("item3", sr)]
    fzf.selected = ["item1", "item2", "item3"]
    fzf.kb_deselect_all()
    assert fzf.selected == []
    fzf.kb_move_items_cursor_relative(1)
    fzf.kb_deselect_all()
    assert fzf.selected == []
    fzf.selected = ["item1", "item2", "item3", "item4", "item5"]
    fzf.kb_deselect_all()
    assert fzf.selected == ["item4", "item5"]


def test_kb_reset_query():
    sr = ScoringResult("", "")
    fzf = FuzzyFinder(query="my initial query")
    fzf.filtered = [("item1", sr), ("item2", sr), ("item3", sr)]
    assert fzf.query == "my initial query"
    fzf.kb_move_items_cursor_absolute(2)
    assert fzf.cursor_items == 2
    assert fzf.cursor_query == 16
    fzf.kb_reset_query()
    assert fzf.query == ""
    assert fzf.cursor_items == 0
    assert fzf.cursor_query == 0
    fzf.kb_move_items_cursor_absolute(2)
    assert fzf.cursor_items == 2
    fzf.kb_reset_query()
    assert fzf.cursor_items == 2


def test_kb_add_to_query():
    sr = ScoringResult("", "")
    fzf = FuzzyFinder(query="my initial query")
    fzf.filtered = [("item1", sr), ("item2", sr), ("item3", sr)]
    assert fzf.query == "my initial query"
    fzf.kb_move_items_cursor_absolute(2)
    fzf.kb_move_query_cursor_absolute(3)
    assert fzf.cursor_items == 2
    assert fzf.cursor_query == 3
    # add after cursor
    fzf.kb_add_to_query("a", 4)
    assert fzf.query == "my ianitial query"
    assert fzf.cursor_items == 0
    assert fzf.cursor_query == 3
    # add before cursor
    fzf.kb_add_to_query("b", 3)
    assert fzf.query == "my bianitial query"
    assert fzf.cursor_items == 0
    assert fzf.cursor_query == 4
    # add at the end (index = -1)
    fzf.kb_add_to_query("c")
    assert fzf.query == "my bianitial queryc"
    assert fzf.cursor_items == 0
    assert fzf.cursor_query == 4
    # add at the end
    fzf.kb_add_to_query("d", -1)
    assert fzf.query == "my bianitial querycd"
    assert fzf.cursor_items == 0
    assert fzf.cursor_query == 4
    # add at the end
    fzf.kb_add_to_query("e", len(fzf.query))
    assert fzf.query == "my bianitial querycde"
    assert fzf.cursor_items == 0
    assert fzf.cursor_query == 4
    # add at the beginning (longer text)
    fzf.kb_add_to_query("hello world ", 0)
    assert fzf.query == "hello world my bianitial querycde"
    assert fzf.cursor_items == 0
    assert fzf.cursor_query == 16
    # add nothing before and after cursor
    # (which should not reset the items cursor either)
    fzf.kb_move_items_cursor_absolute(2)
    assert fzf.cursor_items == 2
    fzf.kb_add_to_query("", 10)
    assert fzf.query == "hello world my bianitial querycde"
    assert fzf.cursor_items == 2
    assert fzf.cursor_query == 16
    fzf.kb_add_to_query("", 20)
    assert fzf.query == "hello world my bianitial querycde"
    assert fzf.cursor_items == 2
    assert fzf.cursor_query == 16
    # it should raise CursesFzfIndexOutOfBounds,
    # which inherrits from CursesFzfAssertion
    with pytest.raises(CursesFzfAssertion):
        fzf.kb_add_to_query("x", -99)
    with pytest.raises(CursesFzfIndexOutOfBounds):
        fzf.kb_add_to_query("x", 99)


def test_kb_add_to_query_cursor():
    sr = ScoringResult("", "")
    fzf = FuzzyFinder(query="my initial query")
    fzf.filtered = [("item1", sr), ("item2", sr), ("item3", sr)]
    assert fzf.query == "my initial query"
    fzf.kb_move_items_cursor_absolute(2)
    assert fzf.cursor_items == 2
    assert fzf.cursor_query == 16
    fzf.kb_add_to_query_cursor("a")
    assert fzf.query == "my initial querya"
    assert fzf.cursor_items == 0
    assert fzf.cursor_query == 17
    fzf.kb_move_query_cursor_absolute(3)
    fzf.kb_add_to_query_cursor("foo")
    assert fzf.query == "my fooinitial querya"
    assert fzf.cursor_items == 0
    assert fzf.cursor_query == 6
    fzf.kb_move_query_cursor_absolute(0)
    fzf.kb_add_to_query_cursor("bar")
    assert fzf.query == "barmy fooinitial querya"
    assert fzf.cursor_items == 0
    assert fzf.cursor_query == 3


def test_kb_remove_from_query():
    sr = ScoringResult("", "")
    fzf = FuzzyFinder(query="my initial query")
    fzf.filtered = [("item1", sr), ("item2", sr), ("item3", sr)]
    assert fzf.query == "my initial query"
    fzf.kb_move_items_cursor_absolute(2)
    assert fzf.cursor_items == 2
    assert fzf.cursor_query == 16
    # it should raise CursesFzfIndexOutOfBounds,
    # which inherrits from CursesFzfAssertion
    with pytest.raises(CursesFzfAssertion):
        fzf.kb_remove_from_query(-99)
    with pytest.raises(CursesFzfIndexOutOfBounds):
        fzf.kb_remove_from_query(99)
    with pytest.raises(CursesFzfAssertion):
        fzf.kb_remove_from_query(2, -5)
    with pytest.raises(CursesFzfIndexOutOfBounds):
        fzf.kb_remove_from_query(2, -1)
    assert fzf.query == "my initial query"
    assert fzf.cursor_items == 2
    assert fzf.cursor_query == 16
    # remove nothing is valid but doesn't change anything
    fzf.kb_remove_from_query(2, 0)
    assert fzf.query == "my initial query"
    assert fzf.cursor_items == 2
    assert fzf.cursor_query == 16
    # remove last char
    # cursor > index (at end of string) -> - length
    fzf.kb_remove_from_query(len(fzf.query)-1)
    assert fzf.query == "my initial quer"
    assert fzf.cursor_items == 0
    assert fzf.cursor_query == 15
    # remove with positive index
    # cursor > index (at end of string) -> - length
    fzf.kb_remove_from_query(4)
    assert fzf.query == "my iitial quer"
    assert fzf.cursor_items == 0
    assert fzf.cursor_query == 14
    # remove with length and positive index
    # cursor > index, but in removed part -> set to index
    fzf.kb_move_query_cursor_absolute(5)
    assert fzf.cursor_query == 5
    fzf.kb_remove_from_query(4, 4)
    assert fzf.query == "my il quer"
    assert fzf.cursor_items == 0
    assert fzf.cursor_query == 4
    # remove last char (index=-1)
    # cursor < index -> keep
    fzf.kb_move_query_cursor_absolute(2)
    assert fzf.cursor_query == 2
    fzf.kb_remove_from_query()
    assert fzf.query == "my il que"
    assert fzf.cursor_items == 0
    assert fzf.cursor_query == 2
    # remove last char
    # curser at index -> keep
    fzf.kb_move_query_cursor_absolute(8)
    assert fzf.cursor_query == 8
    fzf.kb_remove_from_query(-1)
    assert fzf.query == "my il qu"
    assert fzf.cursor_items == 0
    assert fzf.cursor_query == 8
    # remove with negative index
    # cusor > index and end of string, but in removed part -> set to index
    fzf.kb_remove_from_query(-3, 3)
    assert fzf.query == "my il"
    assert fzf.cursor_items == 0
    assert fzf.cursor_query == 5
    # remove with too big length (which is ok)
    # cursor > index and length too big, but in removed part -> set to index
    fzf.kb_remove_from_query(4, 99)
    assert fzf.query == "my i"
    assert fzf.cursor_items == 0
    assert fzf.cursor_query == 4
    # remove from cursor position should not change cursor position
    # curser at index -> keep
    fzf.kb_move_query_cursor_absolute(2)
    fzf.kb_remove_from_query(2, 99)
    assert fzf.query == "my"
    assert fzf.cursor_items == 0
    assert fzf.cursor_query == 2


def test_kb_remove_from_query_cursor():
    sr = ScoringResult("", "")
    fzf = FuzzyFinder(query="my initial query")
    fzf.filtered = [("item1", sr), ("item2", sr), ("item3", sr)]
    assert fzf.query == "my initial query"
    fzf.kb_move_items_cursor_absolute(2)
    assert fzf.cursor_items == 2
    assert fzf.cursor_query == 16
    # remove at the end of query
    fzf.kb_remove_from_query_cursor()
    assert fzf.query == "my initial quer"
    assert fzf.cursor_items == 0
    assert fzf.cursor_query == 15
    # remove at position in the middle of the query
    fzf.kb_move_query_cursor_absolute(5)
    assert fzf.cursor_query == 5
    fzf.kb_remove_from_query_cursor()
    assert fzf.query == "my iitial quer"
    assert fzf.cursor_items == 0
    assert fzf.cursor_query == 4
    # remove at the beginning of the query (don't actually remove anything)
    fzf.kb_move_query_cursor_absolute(0)
    fzf.kb_move_items_cursor_absolute(2)
    assert fzf.cursor_query == 0
    fzf.kb_remove_from_query_cursor()
    assert fzf.query == "my iitial quer"
    assert fzf.cursor_items == 2
    assert fzf.cursor_query == 0
    # remove with before=False should remove the character at the cursor
    fzf.kb_remove_from_query_cursor(False)
    assert fzf.query == "y iitial quer"
    assert fzf.cursor_items == 0
    assert fzf.cursor_query == 0
    # remove with before=False should remove the character at the cursor
    fzf.kb_move_query_cursor_absolute(5)
    assert fzf.cursor_query == 5
    fzf.kb_remove_from_query_cursor(False)
    assert fzf.query == "y iital quer"
    assert fzf.cursor_items == 0
    assert fzf.cursor_query == 5
    fzf.kb_move_query_cursor_absolute(12)
    fzf.kb_move_items_cursor_absolute(2)
    assert fzf.cursor_items == 2
    assert fzf.cursor_query == 12
    # remove with before=False should do nothing if the
    # cursor is at the end of the query
    fzf.kb_remove_from_query_cursor(False)
    assert fzf.query == "y iital quer"
    assert fzf.cursor_items == 2
    assert fzf.cursor_query == 12


def test_keymap():
    sr = ScoringResult("", "")
    fzf = FuzzyFinder(multi=True, query="my initial query", page_size=2)
    fzf.filtered = [("item1", sr), ("item2", sr), ("item3", sr), ("item4", sr),
                    ("item5", sr), ("item6", sr), ("item7", sr)]
    assert isinstance(fzf.keymap, dict)
    assert curses.KEY_UP in fzf.keymap  # Up arrow
    fzf._cursor_items = 3
    fzf.keymap[curses.KEY_UP]()
    assert fzf.cursor_items == 2
    assert curses.KEY_DOWN in fzf.keymap  # Down arrow
    fzf._cursor_items = 3
    fzf.keymap[curses.KEY_DOWN]()
    assert fzf.cursor_items == 4
    assert curses.KEY_PPAGE in fzf.keymap  # Page Up
    fzf._cursor_items = 3
    fzf.keymap[curses.KEY_PPAGE]()
    assert fzf.cursor_items == 1
    assert curses.KEY_NPAGE in fzf.keymap  # Page Down
    fzf._cursor_items = 3
    fzf.keymap[curses.KEY_NPAGE]()
    assert fzf.cursor_items == 5
    assert curses.KEY_HOME in fzf.keymap  # Home
    fzf._cursor_items = 3
    fzf.keymap[curses.KEY_HOME]()
    assert fzf.cursor_items == 0
    assert curses.KEY_END in fzf.keymap  # End
    fzf._cursor_items = 3
    fzf.keymap[curses.KEY_END]()
    assert fzf.cursor_items == 6
    assert curses.KEY_LEFT in fzf.keymap  # Left arrow
    fzf._cursor_query = 5
    fzf.keymap[curses.KEY_LEFT]()
    assert fzf.cursor_query == 4
    assert curses.KEY_RIGHT in fzf.keymap  # Right arrow
    fzf._cursor_query = 5
    fzf.keymap[curses.KEY_RIGHT]()
    assert fzf.cursor_query == 6
    assert curses.KEY_BACKSPACE in fzf.keymap  # Backspace
    fzf._cursor_query = 5
    fzf.keymap[curses.KEY_BACKSPACE]()
    assert fzf.query == "my iitial query"
    assert fzf.cursor_query == 4
    assert curses.KEY_DC in fzf.keymap  # Delete
    fzf._cursor_query = 5
    fzf.keymap[curses.KEY_DC]()
    assert fzf.query == "my iiial query"
    assert fzf.cursor_query == 5
    assert 11 in fzf.keymap  # Ctrl+K
    fzf._cursor_query = 5
    fzf.keymap[11]()
    assert fzf.query == ""
    assert fzf.cursor_query == 0
    assert 27 in fzf.keymap  # Esc
    with pytest.raises(CursesFzfAborted):
        fzf.keymap[27]()
    assert curses.KEY_ENTER in fzf.keymap  # Enter
    assert fzf.return_selection_now == False  # noqa: E712
    fzf.keymap[curses.KEY_ENTER]()
    assert fzf.return_selection_now == True  # noqa: E712
    assert 9 in fzf.keymap  # Tab
    fzf._cursor_items = 1
    fzf.selected = []
    fzf.keymap[9]()
    assert fzf.selected == ["item2"]
    assert 1 in fzf.keymap  # Ctrl+A
    fzf.selected = []
    fzf.keymap[1]()
    assert fzf.selected == ["item1", "item2", "item3", "item4", "item5", "item6", "item7"]
    assert 24 in fzf.keymap  # Ctrl+X
    fzf.keymap[24]()
    assert fzf.selected == []
    assert 16 in fzf.keymap  # Ctrl+P
    assert fzf.show_preview == True  # noqa: E712
    fzf.keymap[16]()
    assert fzf.show_preview == False  # noqa: E712
    assert curses.KEY_F1 in fzf.keymap  # F1
    assert fzf.keymap[curses.KEY_F1] == fzf.kb_show_help


def test_handle_input():
    sr = ScoringResult("", "")
    fzf = FuzzyFinder(query="abc")
    fzf.filtered = [("item1", sr), ("item2", sr), ("item3", sr)]
    assert fzf.cursor_items == 0
    assert fzf.cursor_query == 3
    assert fzf.query == "abc"
    fzf._handle_input(curses.KEY_DOWN)
    assert fzf.cursor_items == 1
    fzf._handle_input('x')
    assert fzf.query == "abcx"
    assert fzf.cursor_query == 4
    assert fzf.cursor_items == 0
    fzf._handle_input(curses.KEY_LEFT)
    assert fzf.cursor_query == 3
    fzf._handle_input('ö')
    assert fzf.query == "abcöx"
    assert fzf.cursor_query == 4
    assert fzf.cursor_items == 0
    fzf._handle_input(curses.KEY_BACKSPACE)
    assert fzf.query == "abcx"
    assert fzf.cursor_query == 3
    assert fzf.cursor_items == 0
    fzf._handle_input(curses.KEY_DC)
    assert fzf.query == "abc"
    assert fzf.cursor_query == 3
    assert fzf.cursor_items == 0


def test_cursor_items():
    sr = ScoringResult("", "")
    fzf = FuzzyFinder()
    fzf.filtered = [("item1", sr), ("item2", sr), ("item3", sr)]
    assert fzf.cursor_items == 0
    fzf.kb_move_items_cursor_absolute(1)
    assert fzf.cursor_items == 1
    # cursor_items should not be settable directly, but only through the
    # kb_move_items_cursor_absolute and kb_move_items_cursor_relative methods
    with pytest.raises(AttributeError):
        fzf.cursor_items = 2  # type: ignore


def test_cursor_query():
    fzf = FuzzyFinder(query="my initial query")
    assert fzf.cursor_query == 16
    fzf.kb_move_query_cursor_absolute(4)
    assert fzf.cursor_query == 4
    # cursor_items should not be settable directly, but only through the
    # kb_move_items_cursor_absolute and kb_move_items_cursor_relative methods
    with pytest.raises(AttributeError):
        fzf.cursor_query = 2  # type: ignore


def test_query():
    sr = ScoringResult("", "")
    fzf = FuzzyFinder(query="my initial query")
    fzf.filtered = [("item1", sr), ("item2", sr), ("item3", sr)]
    fzf.kb_move_items_cursor_absolute(2)
    assert fzf.cursor_items == 2
    assert fzf.cursor_query == 16
    assert fzf.query == "my initial query"
    # getter should not effect any cursor position
    assert fzf.cursor_items == 2
    assert fzf.cursor_query == 16
    fzf.query = "new query"
    assert fzf.query == "new query"
    # setter should ...
    assert fzf.cursor_items == 0
    assert fzf.cursor_query == 9


def test_calculate_filtered():
    fzf = FuzzyFinder(query="app")
    fzf.all_items = ["apple", "banana", "orange"]
    fzf.calculate_filtered()
    assert len(fzf.filtered) == 1
    assert fzf.filtered[0][0] == "apple"
    assert fzf.filtered[0][1].score > 0


def test_calculate_preselection():
    sr = ScoringResult("", "")
    fzf = FuzzyFinder(multi=True, preselect=lambda item, score: item == "item2")
    fzf.filtered = [("item1", sr), ("item2", sr), ("item3", sr)]
    fzf._calculate_preselection()
    assert fzf.selected == ["item2"]


def test_get_return_value():
    sr = ScoringResult("", "")
    fzf_multi = FuzzyFinder(multi=True)
    fzf_multi.filtered = [("item1", sr), ("item2", sr), ("item3", sr)]
    # multi empty selection
    assert fzf_multi._get_return_value() == []
    # multi non-empty selection
    fzf_multi.selected = ["item1", "item2"]
    assert fzf_multi._get_return_value() == ["item1", "item2"]
    fzf_single = FuzzyFinder(multi=False)
    fzf_single.filtered = [("item1", sr), ("item2", sr), ("item3", sr)]
    # sinple empty selection should return the currently highlighted item
    fzf_single.kb_move_items_cursor_absolute(1)
    assert fzf_single._get_return_value() == ["item2"]
    # single non-empty (manipulated) selection should return the selected items list
    fzf_single.selected = ["item3"]
    assert fzf_single._get_return_value() == ["item3"]


def test_autoreturn():
    sr = ScoringResult("", "")
    fzf_multi = FuzzyFinder(multi=True, autoreturn=3)
    fzf_multi.filtered = [("item1", sr), ("item2", sr), ("item3", sr)]
    assert fzf_multi._autoreturn() == ["item1", "item2", "item3"]
    fzf_multi.filtered = [("item1", sr), ("item2", sr)]
    assert fzf_multi._autoreturn() is None
    fzf_multi.filtered = [("item1", sr), ("item2", sr), ("item3", sr), ("item4", sr)]
    assert fzf_multi._autoreturn() is None
    fzf_single = FuzzyFinder(multi=False, autoreturn=3)
    fzf_single.filtered = [("item1", sr), ("item2", sr), ("item3", sr)]
    assert fzf_single._autoreturn() is None
    fzf_single.filtered = [("item1", sr)]
    assert fzf_single._autoreturn() == ["item1"]


def test_render_query():
    # test short query with cursor at the end
    mock_stdscr = MagicMock(spec=curses.window)
    fzf = FuzzyFinder(query="myq")
    fzf.stdscr = mock_stdscr
    with patch('curses.color_pair') as mock_color_pair:
        mock_color_pair.side_effect = lambda x: x   # return color pair ID directly
        fzf._render_query(10)
    mock_stdscr.addstr.assert_has_calls([
        call(0, 4, 'm', 33),
        call(0, 5, 'y', 33),
        call(0, 6, 'q', 33),
        call(0, 7, ' ', 47),
    ], any_order=False)
    # test too long query with cursor in the middle
    mock_stdscr = MagicMock(spec=curses.window)
    fzf = FuzzyFinder(query="myquery")
    fzf.stdscr = mock_stdscr
    fzf.kb_move_query_cursor_absolute(2)
    with patch('curses.color_pair') as mock_color_pair:
        mock_color_pair.side_effect = lambda x: x   # return color pair ID directly
        fzf._render_query(10)
    mock_stdscr.addstr.assert_has_calls([
        call(0, 4, 'm', 33),
        call(0, 5, 'y', 33),
        call(0, 6, 'q', 47),
        call(0, 7, 'u', 33),
    ], any_order=False)


def test_render_no_match():
    # enough width for text
    mock_stdscr = MagicMock(spec=curses.window)
    fzf = FuzzyFinder()
    fzf.stdscr = mock_stdscr
    fzf.filtered = []
    with patch('curses.color_pair') as mock_color_pair:
        mock_color_pair.side_effect = lambda x: x   # return color pair ID directly
        fzf._render_no_match(24)
    mock_stdscr.addstr.assert_has_calls([
        call(3, 4, 'No matching items!', 31),
    ], any_order=False)
    # not enough width for text
    mock_stdscr = MagicMock(spec=curses.window)
    fzf = FuzzyFinder()
    fzf.stdscr = mock_stdscr
    fzf.filtered = []
    with patch('curses.color_pair') as mock_color_pair:
        mock_color_pair.side_effect = lambda x: x   # return color pair ID directly
        fzf._render_no_match(10)
    mock_stdscr.addstr.assert_has_calls([
        call(3, 4, 'No m', 31),
    ], any_order=False)


def test_render_viewport():
    sr = ScoringResult("", "")
    sr.add_match(1, "te", 5)
    mock_stdscr = MagicMock(spec=curses.window)
    fzf = FuzzyFinder()
    fzf.stdscr = mock_stdscr
    fzf.filtered = [("item1", sr), ("item2", sr)]
    with patch('curses.color_pair') as mock_color_pair:
        mock_color_pair.side_effect = lambda x: x   # return color pair ID directly
        fzf._render_viewport(24, 24)
    mock_stdscr.addstr.assert_has_calls([
        call(3, 2, '   ', 47),
        call(3, 5, 'i', 47),
        call(3, 6, 't', 46),
        call(3, 7, 'e', 46),
        call(3, 8, 'm', 47),
        call(3, 9, '1', 47),
        call(4, 2, '   ', 37),
        call(4, 5, 'i', 37),
        call(4, 6, 't', 46),
        call(4, 7, 'e', 46),
        call(4, 8, 'm', 37),
        call(4, 9, '2', 37)
    ], any_order=False)
    # too small window
    mock_stdscr = MagicMock(spec=curses.window)
    fzf = FuzzyFinder()
    fzf.stdscr = mock_stdscr
    fzf.filtered = [("item1", sr), ("item2", sr)]
    with patch('curses.color_pair') as mock_color_pair:
        mock_color_pair.side_effect = lambda x: x   # return color pair ID directly
        fzf._render_viewport(7, 12)
    mock_stdscr.addstr.assert_has_calls([
        call(3, 2, '   ', 47),
        call(3, 5, 'i', 47),
        call(3, 6, 't', 46),
        call(3, 6, '…', 47)
    ], any_order=False)
    # multi
    mock_stdscr = MagicMock(spec=curses.window)
    fzf = FuzzyFinder(multi=True)
    fzf.stdscr = mock_stdscr
    fzf.filtered = [("item1", sr), ("item2", sr)]
    fzf.selected = ["item2", "item1"]
    with patch('curses.color_pair') as mock_color_pair:
        mock_color_pair.side_effect = lambda x: x   # return color pair ID directly
        fzf._render_viewport(24, 24)
    mock_stdscr.addstr.assert_has_calls([
        call(3, 2, '✅ ', 42),
        call(3, 5, 'i', 42),
        call(3, 6, 't', 46),
        call(3, 7, 'e', 46),
        call(3, 8, 'm', 42),
        call(3, 9, '1', 42),
        call(4, 2, '✅ ', 32),
        call(4, 5, 'i', 32),
        call(4, 6, 't', 46),
        call(4, 7, 'e', 46),
        call(4, 8, 'm', 32),
        call(4, 9, '2', 32)
    ], any_order=False)
    # exception
    mock_stdscr = MagicMock(spec=curses.window)
    fzf = FuzzyFinder(display=lambda x: f"{x}\nfoo")
    fzf.stdscr = mock_stdscr
    fzf.filtered = [("item1", sr), ("item2", sr)]
    with patch('curses.color_pair') as mock_color_pair:
        mock_color_pair.side_effect = lambda x: x   # return color pair ID directly
        with pytest.raises(CursesFzfAssertion):
            fzf._render_viewport(24, 24)


def test_render_preview_noop():
    fzf = FuzzyFinder(preview=lambda w, c, i, s: f"{i}\n{s.score}")
    # stdscr not set
    assert fzf._render_preview(20, 50) is None
    # deactivate preview if window gets too small
    mock_stdscr = MagicMock(spec=curses.window)
    fzf.stdscr = mock_stdscr
    fzf.show_preview = True
    assert fzf._render_preview(6, 50) is None
    assert fzf.show_preview == False  # noqa: E712
    fzf.show_preview = True
    assert fzf._render_preview(20, 29) is None
    assert fzf.show_preview == False  # noqa: E712
    # return if show_preview is False
    assert fzf._render_preview(20, 50) is None
    # terutn if no preview function is set
    fzf = FuzzyFinder()
    mock_stdscr = MagicMock(spec=curses.window)
    fzf.stdscr = mock_stdscr
    assert fzf._render_preview(20, 50) is None


def test_render_preview():
    sr = ScoringResult("", "")
    fzf = FuzzyFinder(preview=lambda w, c, i, s: f"{i}\n{s.score}\nfoo\nbar")
    fzf.filtered = [("item1", sr), ("item2", sr)]
    mock_stdscr = MagicMock(spec=curses.window)
    mock_sub_win = MagicMock(spec=curses.window)
    mock_sub_win.getmaxyx.return_value = (6, 12)
    fzf.stdscr = mock_stdscr
    fzf.show_preview = True
    with patch('curses.color_pair') as mock_color_pair, \
         patch('curses.newwin') as mock_newwin:
        mock_color_pair.side_effect = lambda x: x   # return color pair ID directly
        mock_newwin.return_value = mock_sub_win
        rwin = fzf._render_preview(10, 30)
        assert rwin == mock_sub_win
        mock_newwin.assert_called_with(6, 12, 2, 16)
        mock_sub_win.box.assert_called_once()
        mock_sub_win.getmaxyx.assert_called_once()
        mock_sub_win.addstr.assert_has_calls([
            call(0, 2, ' PREVIEW ', 33),
            call(2, 4, 'item1', 37),
            call(3, 4, '0', 37),
        ], any_order=False)
        # should have been truncated
        assert call(4, 4, "foo", 37) not in mock_sub_win.addstr.call_args_list
        assert call(5, 4, "bar", 37) not in mock_sub_win.addstr.call_args_list


def test_find():
    fzf = FuzzyFinder(title="my initial title", query="my initial query")
    assert fzf.all_items == []
    assert fzf.title == "my initial title"
    assert fzf.query == "my initial query"
    itemlist = ["item1", "item2", "item3"]
    # test using object parameters
    with patch('curses.wrapper') as mock_wrapper:
        fzf.find(itemlist)
        mock_wrapper.assert_called_once()
        assert fzf.all_items == itemlist
        assert fzf.title == "my initial title"
        assert fzf.query == "my initial query"
    # test with local override parameters
    with patch('curses.wrapper') as mock_wrapper:
        fzf.find(itemlist, title="new title", query="new query")
        mock_wrapper.assert_called_once()
        assert fzf.all_items == itemlist
        assert fzf.title == "new title"
        assert fzf.query == "new query"
    # test Ctrl + C (KeyboardInterrupt) handling
    with patch.object(curses, "wrapper", side_effect=KeyboardInterrupt):
        with pytest.raises(CursesFzfAborted):
            fzf.find(itemlist)



import pytest
from curses_fzf import Color, ColorTheme
from curses_fzf.colors import _init_curses
from unittest.mock import patch
import curses


@pytest.fixture
def default():
    return ColorTheme()


@pytest.fixture
def custom():
    return ColorTheme(text=Color.RED, window_title=Color.GREEN, no_match=Color.BLUE, query=Color.MAGENTA,
                      footer=Color.CYAN, selected=Color.YELLOW, cursor=Color.BLACK_ON_RED,
                      cursor_selected=Color.WHITE_ON_BLUE, highlight=Color.WHITE_ON_MAGENTA)


def test_scoringresult_init(default, custom):
    assert default.text == Color.WHITE
    assert default.window_title == Color.YELLOW
    assert default.no_match == Color.RED
    assert default.query == Color.YELLOW
    assert default.footer == Color.YELLOW
    assert default.selected == Color.GREEN
    assert default.cursor == Color.BLACK_ON_WHITE
    assert default.cursor_selected == Color.BLACK_ON_GREEN
    assert default.highlight == Color.BLACK_ON_CYAN
    assert custom.text == Color.RED
    assert custom.window_title == Color.GREEN
    assert custom.no_match == Color.BLUE
    assert custom.query == Color.MAGENTA
    assert custom.footer == Color.CYAN
    assert custom.selected == Color.YELLOW
    assert custom.cursor == Color.BLACK_ON_RED
    assert custom.cursor_selected == Color.WHITE_ON_BLUE
    assert custom.highlight == Color.WHITE_ON_MAGENTA


def test_init_curses():
    with patch("curses.curs_set") as mock_curs_set, \
            patch("curses.start_color") as mock_start_color, \
            patch("curses.use_default_colors") as mock_use_default_colors, \
            patch("curses.init_pair") as mock_init_pair:
        _init_curses()
        mock_curs_set.assert_called_with(0)
        mock_start_color.assert_called_once()
        mock_use_default_colors.assert_called_once()
        assert mock_init_pair.call_count == 22
        mock_init_pair.assert_any_call(Color.BLACK, curses.COLOR_BLACK, -1)
        mock_init_pair.assert_any_call(Color.RED, curses.COLOR_RED, -1)
        mock_init_pair.assert_any_call(Color.GREEN, curses.COLOR_GREEN, -1)
        mock_init_pair.assert_any_call(Color.YELLOW, curses.COLOR_YELLOW, -1)
        mock_init_pair.assert_any_call(Color.BLUE, curses.COLOR_BLUE, -1)
        mock_init_pair.assert_any_call(Color.MAGENTA, curses.COLOR_MAGENTA, -1)
        mock_init_pair.assert_any_call(Color.CYAN, curses.COLOR_CYAN, -1)
        mock_init_pair.assert_any_call(Color.WHITE, curses.COLOR_WHITE, -1)
        mock_init_pair.assert_any_call(Color.BLACK_ON_RED, curses.COLOR_WHITE, curses.COLOR_RED)
        mock_init_pair.assert_any_call(Color.BLACK_ON_GREEN, curses.COLOR_BLACK, curses.COLOR_GREEN)
        mock_init_pair.assert_any_call(Color.BLACK_ON_YELLOW, curses.COLOR_BLACK, curses.COLOR_YELLOW)
        mock_init_pair.assert_any_call(Color.BLACK_ON_BLUE, curses.COLOR_BLACK, curses.COLOR_BLUE)
        mock_init_pair.assert_any_call(Color.BLACK_ON_MAGENTA, curses.COLOR_BLACK, curses.COLOR_MAGENTA)
        mock_init_pair.assert_any_call(Color.BLACK_ON_CYAN, curses.COLOR_BLACK, curses.COLOR_CYAN)
        mock_init_pair.assert_any_call(Color.BLACK_ON_WHITE, curses.COLOR_BLACK, curses.COLOR_WHITE)
        mock_init_pair.assert_any_call(Color.WHITE_ON_BLACK, curses.COLOR_WHITE, curses.COLOR_BLACK)
        mock_init_pair.assert_any_call(Color.WHITE_ON_RED, curses.COLOR_WHITE, curses.COLOR_RED)
        mock_init_pair.assert_any_call(Color.WHITE_ON_GREEN, curses.COLOR_WHITE, curses.COLOR_GREEN)
        mock_init_pair.assert_any_call(Color.WHITE_ON_YELLOW, curses.COLOR_WHITE, curses.COLOR_YELLOW)
        mock_init_pair.assert_any_call(Color.WHITE_ON_BLUE, curses.COLOR_WHITE, curses.COLOR_BLUE)
        mock_init_pair.assert_any_call(Color.WHITE_ON_MAGENTA, curses.COLOR_WHITE, curses.COLOR_MAGENTA)
        mock_init_pair.assert_any_call(Color.WHITE_ON_CYAN, curses.COLOR_WHITE, curses.COLOR_CYAN)



class CursesFzfException(Exception):
    """
    Base exception class for :class:`~curses_fzf.FuzzyFinder`.
    You can catch this exception to handle all exceptions raised
    by this module.
    """
    pass


class CursesFzfAborted(CursesFzfException):
    """
    Exception raised when :class:`~curses_fzf.FuzzyFinder` is aborted by the user
    (e.g. by pressing :kbd:`Ctrl + C` or :kbd:`ESC`).
    """
    pass


class CursesFzfAssertion(CursesFzfException):
    """
    Exception raised when an assertion fails in :class:`~curses_fzf.FuzzyFinder`.
    E.g. if the given :meth:`~curses_fzf.FuzzyFinder.display` function returns multi-line text.
    """
    pass


class CursesFzfIndexOutOfBounds(CursesFzfAssertion):
    """
    Exception raised when an index violation occurs.
    E.g. if :class:`~curses_fzf.FuzzyFinder`'s user tries to access the :attr:`~curses_fzf.FuzzyFinder.query`
    at an invalid index using :meth:`~curses_fzf.FuzzyFinder.kb_add_to_query`.
    """
    pass



#!/usr/bin/env python3
import curses
from typing import Any
from curses_fzf import FuzzyFinder, ScoringResult, ColorTheme, CursesFzfAborted
import yaml


def yaml_preview(preview_window: curses.window, color_theme: ColorTheme, item: Any, result: ScoringResult) -> str:
    """
    A preview function using the simple text return mechanism.
    """
    return yaml.safe_dump(item)


def main() -> None:
    """
    Example: dict items with simple preview and preselect
    """
    fzf = FuzzyFinder(
        # fuzzyfind data allowing selection of multiple items
        multi=True,
        # display dict items by "name" key
        display=lambda item: item.get("name"),
        # preselect every item with less than 400 calories
        preselect=lambda item, result: item.get("calories", 0) < 400,
        # display preview as yaml representation of our items
        preview=yaml_preview,
        # reduce page size since we only have a short list
        page_size=5
    )
    try:
        result = fzf.find(DATA)
    except CursesFzfAborted:
        print("Fuzzy finder aborted by user.")
        return
    # in multi selection mode it is possible that the result is an empty list if the user
    # accepted with Enter without selecting any item
    for item in result:
        print(item.get("name"))


DATA = [
    {"id": 1, "name": "apple pie", "tags": ["sweet", "fruit"], "calories": 450, "price": 4.99, "available": True},
    {"id": 2, "name": "banana split", "tags": ["sweet", "cold"], "calories": 680, "price": 6.50, "available": True},
    {"id": 3, "name": "beef stroganoff", "tags": ["savory", "meat"], "calories": 620, "price": 12.99,
     "available": True},
    {"id": 4, "name": "vegetable stir-fry", "tags": ["vegan", "quick"], "calories": 320, "price": 8.75,
     "available": True},
    {"id": 5, "name": "chocolate lava cake", "tags": ["sweet", "chocolate"], "calories": 520, "price": 7.25,
     "available": False},
    {"id": 6, "name": "caesar salad", "tags": ["healthy", "salad"], "calories": 280, "price": 9.50, "available": True},
    {"id": 7, "name": "mushroom risotto", "tags": ["vegetarian", "creamy"], "calories": 410, "price": 11.25,
     "available": True},
    {"id": 8, "name": "grilled salmon", "tags": ["fish", "healthy"], "calories": 380, "price": 15.99,
     "available": True},
    {"id": 9, "name": "pasta carbonara", "tags": ["pasta", "creamy"], "calories": 720, "price": 10.75,
     "available": True},
    {"id": 10, "name": "lemon cheesecake", "tags": ["sweet", "citrus"], "calories": 490, "price": 5.99,
     "available": True},
    {"id": 11, "name": "chicken curry", "tags": ["spicy", "indian"], "calories": 550, "price": 13.50,
     "available": False},
    {"id": 12, "name": "french fries", "tags": ["side", "crispy"], "calories": 360, "price": 3.99, "available": True},
    {"id": 13, "name": "tiramisu", "tags": ["coffee", "dessert"], "calories": 460, "price": 6.75, "available": True},
    {"id": 14, "name": "falafel wrap", "tags": ["vegan", "streetfood"], "calories": 420, "price": 7.99,
     "available": True},
    {"id": 15, "name": "steak sandwich", "tags": ["meat", "sandwich"], "calories": 680, "price": 14.50,
     "available": True}
]


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
import re
from curses_fzf import FuzzyFinder, CursesFzfAborted, ScoringResult, ColorTheme, Color


def my_score(query: str, candidate: str) -> ScoringResult:
    """
    Simple demo scoring function that gives a score of 10 for each occurrence of the query in the candidate.
    Adding the position of the match as a match to the ScoringResult, so that it will be highlighted in the interface.
    """
    sr = ScoringResult(query, candidate)
    matches = [m.start() for m in re.finditer(re.escape(query), candidate)]
    if not matches:
        sr.score = 0
    else:
        for m in matches:
            sr.add_match(m, query, 10)
    return sr


def main() -> None:
    """
    Example: custom scoring function and color theme
    """
    theme = ColorTheme(highlight=Color.MAGENTA, query=Color.MAGENTA)
    fzf = FuzzyFinder(score=my_score, color_theme=theme)
    try:
        result = fzf.find(DATA)
    except CursesFzfAborted:
        print("Fuzzy finder aborted by user.")
        return
    print(result[0])


DATA = [
    "Der Ölprinz ist ein Roman von Karl May, der 1897 veröffentlicht wurde.",
    "The quick brown fox jumps over the lazy dog and lands elegantly on the meadow.",
    "In a small town on the edge of the forest lived an old man who told stories.",
    "Modern technology fundamentally changes our daily lives and brings new challenges.",
    "Last night we enjoyed a delicious dinner with fresh salad and grilled fish.",
    "The children played happily in the park while the parents sat on a bench chatting.",
    "During summer, many families travel to the beach to enjoy sun and sea.",
    "The company's CEO announced a major restructuring affecting hundreds of jobs.",
    "In the library, I found an old book on the history of ancient civilizations.",
    "The new film with the famous actor is breaking records at box offices worldwide.",
    "Early in the morning, she jogs through the city park breathing in fresh air deeply.",
    "The recipe for homemade pizza requires flour, tomato sauce, and fresh mozzarella.",
    "During the conference, experts discussed the future of artificial intelligence.",
    "The politician promised infrastructure improvements, but citizens remain skeptical.",
    "At the zoo, we watched the lions lazily lying in the sun and dozing.",
    "The software developers worked until midnight to fix the bug before release.",
    "On the weekend, we're driving to the mountains to hike and experience nature.",
    "The concert in the stadium was unforgettable with fireworks and great lighting.",
    "She studies medicine at university and dreams of a career as a doctor.",
    "The market offered fresh fruit, vegetables, and handmade products from local farmers.",
    "After the rain, a rainbow appeared in the sky, enchanting all onlookers.",
    "The architects designed a modern building with glass facade and green roofs.",
    "During the presentation, he convinced the team with clear arguments and diagrams.",
    "The novel tells the story of a young woman in the Victorian era.",
    "We planned a trip to Italy visiting Rome, Florence, and Venice.",
    "The gym was full of people implementing their New Year's resolutions.",
    "The teacher explained the formula for the area of a circle: pi times radius squared.",
    "In the museum hung paintings by famous artists like Monet and Van Gogh.",
    "Economic growth is forecast at three percent for this year.",
    "He repaired the bicycle with new brakes and a lubricated chain.",
    "The band rehearsed for hours for the festival coming weekend.",
    "In the kitchen, it smelled of fresh bread from the oven and herbs.",
    "The detective solved the mystery through careful observation of the clues.",
    "The students prepared for the math exam with practice problems.",
    "At the airport, she waited impatiently for the delayed flight to Berlin.",
    "The new smartphone has a better camera and longer battery life.",
    "During winter, we enjoy hot chocolate and cozy blankets.",
    "The journalist interviewed the president about current political issues.",
    "In the garden, roses bloomed in all colors, delighting the eye.",
    "The company invested in sustainable energy sources like wind and solar.",
    "After the accident, neighbors helped with first aid and support.",
    "The play addressed love, betrayal, and redemption in dramatic fashion.",
    "We cooked an Indian curry with spices, rice, and yogurt.",
    "The coach motivated the team before the decisive soccer match.",
    "In the workshop, he built a model airplane from wood and glue.",
    "The author signed books for fans after the reading at the bookstore.",
    "The ocean was calm as the ship entered the harbor at sunset.",
    "She learned to play guitar and composed her own songs.",
    "The project failed due to lack of funding and planning.",
    "At the market square, vendors sold handicrafts and regional specialties.",
    "The scientist discovered a new species of bacteria in the lab.",
    "The family celebrated Christmas with gifts and a large festive meal.",
    "He bought a new car with hybrid drive and navigation system.",
    "In the café, we drank espresso and ate croissants for breakfast.",
    "The expedition explored Antarctica with special equipment and sleds.",
    "The song became a hit on the charts and was streamed millions of times.",
    "She decorated the room with flower vases and candles for the party.",
    "The mechanic changed the oil and checked the car's tires.",
    "In the opera, the soprano sang an aria from Mozart's Magic Flute.",
    "The team won the tournament after an exciting final match.",
    "We planted trees in the park to protect the environment.",
    "The chef prepared a menu with appetizer, main course, and dessert.",
    "The tourists photographed the Eiffel Tower lit up at night.",
    "He studied philosophy and debated ethics and morality.",
    "The weather was sunny with blue skies and a light breeze.",
    "In the hospital, she visited her friend after the operation.",
    "The app helped organize appointments and reminders.",
    "The painter created a portrait with oil paints on canvas.",
    "We rode the train through the landscape enjoying the view.",
    "The book contained recipes for vegan dishes and desserts.",
    "The photographer captured the sunrise over the mountains perfectly.",
    "She trained for the marathon with daily runs.",
    "The meeting lasted three hours with presentations and discussions.",
    "In the aquarium, colorful fish swam in crystal-clear water.",
    "The conductor led the orchestra in the symphony.",
    "The house renovation cost more than planned.",
    "We baked a cake with chocolate and nuts for the birthday.",
    "The pilot landed the plane safely despite strong winds.",
    "In the gallery hung abstract artworks by modern artists.",
    "The seminar covered data analysis with Python and Pandas.",
    "She sewed a dress from silk for the wedding.",
    "The fisherman caught salmon in the river with net and bait.",
    "The series aired on TV with new episodes every week.",
    "He built a computer with high-end components himself.",
    "In the restaurant, they served pasta with truffles and wine.",
    "The hike led through forests and over mountain peaks.",
    "The smartphone update fixes security vulnerabilities and improves performance.",
    "We celebrated New Year's Eve with fireworks and champagne.",
    "The lawyer argued brilliantly in court for his client.",
    "In the bakery, it smelled of fresh bread and pastries.",
    "The experiment succeeded and proved the new hypothesis.",
    "She danced a waltz on the dance floor in the ballroom."
]


if __name__ == "__main__":
    main()



from __future__ import annotations

import re
from typing import Optional, Tuple

RE_WORD = re.compile(r"\S+")


class ScoringResult():
    """
    ScoringResult encapsules information and toolage to help with your own scoring functions.

    The only important field is :attr:`~ScoringResult.score`, :class:`~curses_fzf.FuzzyFinder` will take it
    into account when it decides which items to show and in which order.
    If it is ``0`` the element will not be displayed in :attr:`FuzzyFinder.filtered` list.
    Positive values will sort the items from high to low :attr:`~ScoringResult.score`.

    The second field to notice is :attr:`~ScoringResult.matches`, which is a list of tuples containing
    the starting index of all matches inside the :attr:`~ScoringResult.candidate` string along with
    the matched substring.
    If set, this information will be used by :class:`~curses_fzf.FuzzyFinder` to highlight the matched substrings
    in the :attr:`FuzzyFinder.filtered` list of query results.

    The intended way to set those fields is :meth:`~ScoringResult.add_match`.
    It adds the given score to :attr:`ScoringResult.score` and the given position and match as a
    tuple to :attr:`ScoringResult.matches`.
    However feel free to directly manipulate these fields as your scoring logic requires.
    Especially setting :attr:`~ScoringResult.score` to ``0`` is a common way to filter an item out of the results.

    Args:
        query (int): This is the :attr:`FuzzyFinder.query` string entered by the user.
        candidate (int): This is a single item's :meth:`~FuzzyFinder.display` representation
            from the :attr:`FuzzyFinder.all_items` list.
    """

    def __init__(self, query: str, candidate: str) -> None:
        self.query = query
        """
        The original query string parameter as given to the constructor.
        This is the :attr:`FuzzyFinder.query` string entered by the user.
        """

        self.query_lower = query.lower()
        """
        The :attr:`~ScoringResult.query` string converted to lowercase.
        """

        self.query_words_with_index = [(m.group(), m.start()) for m in RE_WORD.finditer(self.query_lower)]
        """
        The :attr:`~ScoringResult.query_lower` string split on whitespaces.
        Each element is a tuple with the word and its starting index in the
        original :attr:`~ScoringResult.query` string.
        """

        self.candidate = candidate
        """
        The original candidate string parameter as given to the constructor.
        This is a single item's :meth:`~FuzzyFinder.display` representation from
        the :attr:`FuzzyFinder.all_items` list.
        """

        self.candidate_lower = candidate.lower()
        """
        The :attr:`~ScoringResult.candidate` string converted to lowercase.
        """

        self.candidate_words_with_index = [(m.group(), m.start()) for m in RE_WORD.finditer(self.candidate_lower)]
        """
        The :attr:`~ScoringResult.candidate_lower` string split on whitespaces.
        Each element is a tuple with the word and its starting index in the
        original :attr:`~ScoringResult.candidate` string.
        """

        self.score = 0
        """
        The resulting fuzzy score.
        This is the field that :class:`~curses_fzf.FuzzyFinder` will take into
        account when it decides which items to show.
        A score of ``0`` will filter the item out of the resulting :attr:`FuzzyFinder.filtered` list.
        Remaining items are sorted from high to low score.
        The :meth:`~ScoringResult.add_match` method is meant to add to score,
        but feel free to directly manipulate this field as your scoring logic requires.
        """

        self.matches = []
        """
        A list of tuples representing each substring match of the :attr:`~ScoringResult.query`
        in the :attr:`~ScoringResult.candidate` string.
        The first tuple entry is the starting index of the match in the original
        :attr:`~ScoringResult.candidate` string.
        The second entry is the matched substring.
        The :meth:`~ScoringResult.add_match` method is meant to add to matches,
        but feel free to directly manipulate this field as your scoring logic requires.
        """

        self._already_matched_words = set()
        """
        A set of indices of candidate words that have already been matched to a query word.
        """

    def __int__(self) -> int:
        return int(self.score)

    def __lt__(self, other: object) -> bool:
        if isinstance(other, ScoringResult):
            return self.score < other.score
        try:
            return self.score < int(other)  # type: ignore[arg-type]
        except Exception:
            return NotImplemented

    def __gt__(self, other: object) -> bool:
        if isinstance(other, ScoringResult):
            return self.score > other.score
        try:
            return self.score > int(other)  # type: ignore[arg-type]
        except Exception:
            return NotImplemented

    def __eq__(self, other: object) -> bool:
        if isinstance(other, ScoringResult):
            return self.score == other.score
        try:
            return self.score == int(other)  # type: ignore[arg-type]
        except Exception:
            return NotImplemented

    def add_match(self, position: int, match: str, score: int) -> None:
        """
        Add position and match as a tuple to :attr:`ScoringResult.matches`.
        The given score is added to :attr:`ScoringResult.score`.

        This is the intended way to add to the :attr:`~ScoringResult.score` and
        keep track of the :attr:`~ScoringResult.matches`, but feel free to
        directly manipulate these fields as your scoring logic requires.
        Especially setting :attr:`~ScoringResult.score` to ``0`` is a common way
        to filter an item out of the results.

        Args:
            position (int): The starting index of the match in the original
                :attr:`~ScoringResult.candidate` string.
            match (str): The matched substring, starting at the given position.
            score (int): This value is added to :attr:`ScoringResult.score`.
        """
        self.score += score
        self.matches.append((position, match))

    def find_best_word_match(self, word: str) -> Optional[Tuple[str, int, int, int]]:
        """
        Find the best match for the given query word among the :attr:`~ScoringResult.candidate_words_with_index`.
        The best match is the one that is closest to a full word (highest percentage of the word matched).
        If multiple matches have the same percentage, the one with the match closer to the word's beginning
        is preferred.

        If a match is found, the index of the matched word in the :attr:`~ScoringResult.candidate` is added to
        :attr:`ScoringResult._already_matched_words` to prevent it from being matched again to another query word.

        Args:
            word (str): The query word (an item from :attr:`~ScoringResult.query_words_with_index`)
                for which to find the best match among :attr:`~ScoringResult.candidate_words_with_index`.

        Returns:
            Optional[Tuple[str, int, int, int]]: A tuple containing:
                - the matched word from :attr:`~ScoringResult.candidate_words_with_index`
                - its position inside :attr:`~ScoringResult.candidate` string
                - the position of the match inside the matched word
                - the percentage of the :attr:`~ScoringResult.candidate` word the :attr:`~ScoringResult.query`
                  word actually matches

            Returns ``None`` if no match was found at all.
        """
        best_match = None
        length_matched_word = -1
        match_position_in_word = -1
        for c_tuple in self.candidate_words_with_index:
            # don't consider the same word twice for different query words
            if c_tuple[1] in self._already_matched_words:
                continue
            pos = c_tuple[0].find(word)
            if pos != -1:
                len_found_word = len(c_tuple[0])
                # the best match is the one that is closest to a full word
                if len_found_word < length_matched_word or \
                        length_matched_word == -1:
                    length_matched_word = len_found_word
                    match_position_in_word = pos
                    best_match = c_tuple
                # if 2 findings have the same length prefer the one with the
                # match closer to the word's beginning
                elif len_found_word == length_matched_word and \
                        pos < match_position_in_word:
                    match_position_in_word = pos
                    best_match = c_tuple
        if best_match is None:
            return None
        # remember the word's index in candidate
        self._already_matched_words.add(best_match[1])
        return (best_match[0], best_match[1], match_position_in_word,
                int(100 * len(word) / length_matched_word))


def scoring_full_words(query: str, candidate: str) -> ScoringResult:
    """
    The :attr:`~curses_fzf.ScoringResult.query` and the :attr:`~curses_fzf.ScoringResult.candidate`
    string both get lowercased and split on whitespaces (see
    :attr:`~curses_fzf.ScoringResult.query_words_with_index` &
    :attr:`~curses_fzf.ScoringResult.candidate_words_with_index`).

    Each query word is supposed to find a unique match among the words of the
    candidate.
    The closer a match is to a full word the higher the score.

    An additional bonus is given, if the match is at the beginning of a word.
    The query words may appear in the candidate in any order, however if the
    original order is found a small bonus will be granted.
    """
    result = ScoringResult(query, candidate)
    # before any query is entered we consider all candidates equal and
    # therefore also keep the original order
    if not query:
        result.score = 100
        return result

    for q_word, q_word_index in result.query_words_with_index:
        best_match = result.find_best_word_match(q_word)
        # all query words need to find a match to keep the candidate
        if best_match is None:
            result.score = 0
            return result

        match_position_in_candidate = best_match[1] + best_match[2]
        # score is the word match percentage multiplied by a bonus if the
        # match starts at the word's beginning
        score = best_match[3]
        if best_match[2] == 0:
            score *= 1.5
        result.add_match(match_position_in_candidate, q_word, int(score))

    # small bonus if all matches are in the exact order of the query
    if all(result.matches[i][0] < result.matches[i+1][0] for i in
            range(len(result.matches) - 1)):
        result.score = int(result.score * 1.2)

    # normalize the score by the number of matches (= number of query words)
    result.score = int(result.score / len(result.matches))
    return result


# TODO provide more scoring functions with different behavior



import curses
from enum import IntEnum


class Color(IntEnum):
    """
    An integer enum with the indices that get registered by :meth:`curses.init_pair`.
    As those indices start at ``30``, you are free to use ``1-29`` to register
    your own additional color pairs.
    """
    BLACK = 30
    """Black text on default terminal background."""
    RED = 31
    """Red text on default terminal background."""
    GREEN = 32
    """Green text on default terminal background."""
    YELLOW = 33
    """Yellow text on default terminal background."""
    BLUE = 34
    """Blue text on default terminal background."""
    MAGENTA = 35
    """Magenta text on default terminal background."""
    CYAN = 36
    """Cyan text on default terminal background."""
    WHITE = 37
    """White text on default terminal background."""
    BLACK_ON_RED = 41
    """Black text on red background."""
    BLACK_ON_GREEN = 42
    """Black text on green background."""
    BLACK_ON_YELLOW = 43
    """Black text on yellow background."""
    BLACK_ON_BLUE = 44
    """Black text on blue background."""
    BLACK_ON_MAGENTA = 45
    """Black text on magenta background."""
    BLACK_ON_CYAN = 46
    """Black text on cyan background."""
    BLACK_ON_WHITE = 47
    """Black text on white background."""
    WHITE_ON_BLACK = 50
    """White text on black background."""
    WHITE_ON_RED = 51
    """White text on red background."""
    WHITE_ON_GREEN = 52
    """White text on green background."""
    WHITE_ON_YELLOW = 53
    """White text on yellow background."""
    WHITE_ON_BLUE = 54
    """White text on blue background."""
    WHITE_ON_MAGENTA = 55
    """White text on magenta background."""
    WHITE_ON_CYAN = 56
    """White text on cyan background."""


class ColorTheme:
    """
    A color theme for the :class:`~curses_fzf.FuzzyFinder`, to allow for easy color customization.
    This is mainly meant to account for color blindness or unusual terminal colors.

    Provide a :class:`ColorTheme` instance to :class:`~curses_fzf.FuzzyFinder`'s
    :attr:`~curses_fzf.FuzzyFinder.color_theme` parameter.

    All color indices are accessible by class attributes of the same name as the
    respective parameter.

    Args:
        text (Color): The normal text (default :attr:`Color.WHITE`).
        window_title (Color): The title in the upper left corner of the main
            and sub windows (default :attr:`Color.YELLOW`).
        no_match (Color): The warning indicating that there are no matched
            items left (default :attr:`Color.RED`).
        query (Color): The query string entered by the user
            (default :attr:`Color.YELLOW`).
        footer (Color): The footer line at the lower end of the main window
            (default :attr:`Color.YELLOW`).
        selected (Color): The text color of selected items in multi-select mode
            (default :attr:`Color.GREEN`).
        cursor (Color): The text color for the line with the cursor
            (default :attr:`Color.BLACK_ON_WHITE`).
        cursor_selected (Color): The text color for the line with the cursor,
            if the item is selected in multi-select mode
            (default :attr:`Color.BLACK_ON_GREEN`).
        highlight (Color): The highlight of the text that was matched by the
            query (default :attr:`Color.BLACK_ON_CYAN`).
    """

    def __init__(self,
                 text: Color = Color.WHITE,
                 window_title: Color = Color.YELLOW,
                 no_match: Color = Color.RED,
                 query: Color = Color.YELLOW,
                 footer: Color = Color.YELLOW,
                 selected: Color = Color.GREEN,
                 cursor: Color = Color.BLACK_ON_WHITE,
                 cursor_selected: Color = Color.BLACK_ON_GREEN,
                 highlight: Color = Color.BLACK_ON_CYAN,
                 ) -> None:
        self.text = text
        self.window_title = window_title
        self.no_match = no_match
        self.query = query
        self.footer = footer
        self.selected = selected
        self.cursor = cursor
        self.cursor_selected = cursor_selected
        self.highlight = highlight


def _init_curses() -> None:
    """
    Setup curses & colors.
    """
    # hide cursor & setup colors
    curses.curs_set(0)
    curses.start_color()
    curses.use_default_colors()
    curses.init_pair(Color.BLACK, curses.COLOR_BLACK, -1)
    curses.init_pair(Color.RED, curses.COLOR_RED, -1)
    curses.init_pair(Color.GREEN, curses.COLOR_GREEN, -1)
    curses.init_pair(Color.YELLOW, curses.COLOR_YELLOW, -1)
    curses.init_pair(Color.BLUE, curses.COLOR_BLUE, -1)
    curses.init_pair(Color.MAGENTA, curses.COLOR_MAGENTA, -1)
    curses.init_pair(Color.CYAN, curses.COLOR_CYAN, -1)
    curses.init_pair(Color.WHITE, curses.COLOR_WHITE, -1)
    curses.init_pair(Color.BLACK_ON_RED, curses.COLOR_WHITE, curses.COLOR_RED)
    curses.init_pair(Color.BLACK_ON_GREEN, curses.COLOR_BLACK, curses.COLOR_GREEN)
    curses.init_pair(Color.BLACK_ON_YELLOW, curses.COLOR_BLACK, curses.COLOR_YELLOW)
    curses.init_pair(Color.BLACK_ON_BLUE, curses.COLOR_BLACK, curses.COLOR_BLUE)
    curses.init_pair(Color.BLACK_ON_MAGENTA, curses.COLOR_BLACK, curses.COLOR_MAGENTA)
    curses.init_pair(Color.BLACK_ON_CYAN, curses.COLOR_BLACK, curses.COLOR_CYAN)
    curses.init_pair(Color.BLACK_ON_WHITE, curses.COLOR_BLACK, curses.COLOR_WHITE)
    curses.init_pair(Color.WHITE_ON_BLACK, curses.COLOR_WHITE, curses.COLOR_BLACK)
    curses.init_pair(Color.WHITE_ON_RED, curses.COLOR_WHITE, curses.COLOR_RED)
    curses.init_pair(Color.WHITE_ON_GREEN, curses.COLOR_WHITE, curses.COLOR_GREEN)
    curses.init_pair(Color.WHITE_ON_YELLOW, curses.COLOR_WHITE, curses.COLOR_YELLOW)
    curses.init_pair(Color.WHITE_ON_BLUE, curses.COLOR_WHITE, curses.COLOR_BLUE)
    curses.init_pair(Color.WHITE_ON_MAGENTA, curses.COLOR_WHITE, curses.COLOR_MAGENTA)
    curses.init_pair(Color.WHITE_ON_CYAN, curses.COLOR_WHITE, curses.COLOR_CYAN)



import curses
from typing import Any, Callable, List, Tuple, Optional, Union

from .colors import ColorTheme, _init_curses
from .help import _help, _base_window
from .errors import CursesFzfAborted, CursesFzfAssertion, CursesFzfIndexOutOfBounds
from .scoring import ScoringResult, scoring_full_words


ITEM_COL_START = 2
SELECTED_MARKER = "✅ "
DESELECTED_MARKER = "   "
CHAR_CONTINUED = "…"
UnicodeKey = Union[int, str]


class FuzzyFinder:
    """
    :class:`~curses_fzf.FuzzyFinder` is the main entry point for the library.
    Use :meth:`~curses_fzf.FuzzyFinder.find` to run the fuzzyfinder on a list of
    items and return the selection.
    It allows the user to filter a list of items based on a
    :attr:`~curses_fzf.FuzzyFinder.query` entered in UI or preseeded in
    constructor.

    Args:
        multi (bool): :attr:`~curses_fzf.FuzzyFinder.multi` selection mode
            determines whether to allow selection of multiple items or not.
            Default is ``False``.
        title (str): The :attr:`~curses_fzf.FuzzyFinder.title` to display in the
            upper left corner of the main :class:`~curses_fzf.FuzzyFinder` window.
            Default is ``"ITEMS"``.
        query (str): The initial :attr:`~curses_fzf.FuzzyFinder.query` to preseed
            the interface with.
            Default is ``""``.
            See :attr:`~curses_fzf.FuzzyFinder.query` for more details.
        display (Callable[[Any], str]): :meth:`~curses_fzf.FuzzyFinder.display` function
            is used to convert an item to a string for display and matching purposes.
            Default is ``lambda item: str(item)``.
            See :meth:`~curses_fzf.FuzzyFinder.display` for more details.
        preselect (Callable[[Any, ScoringResult], bool]): :meth:`~curses_fzf.FuzzyFinder.preselect`
            is a function to determine if an item should be preselected in
            :attr:`~curses_fzf.FuzzyFinder.multi` selection mode based on the item
            and its scoring result.
            Default is a function that always returns ``False``.
            See :meth:`~curses_fzf.FuzzyFinder.preselect` for more details.
        preview (Optional[Callable[[curses.window, ColorTheme, Any, ScoringResult], str]]): Show
            a preview window if :meth:`~curses_fzf.FuzzyFinder.preview` function is provided.
            Default is ``None``.
            See :meth:`~curses_fzf.FuzzyFinder.preview` for more details.
        score (Callable[[str, str], ScoringResult]): The :meth:`~curses_fzf.FuzzyFinder.score`
            function is used to calculate the score of an item from :attr:`~curses_fzf.FuzzyFinder.all_items`
            list based on the current :attr:`~curses_fzf.FuzzyFinder.query` and the
            :meth:`~curses_fzf.FuzzyFinder.display` string of the item.
            Default is :func:`~curses_fzf.scoring_full_words`.
            See :meth:`~curses_fzf.FuzzyFinder.score` for more details.
        color_theme (Optional[ColorTheme]): The :attr:`~curses_fzf.FuzzyFinder.color_theme`
            to use for the interface, if ``None`` was given the default
            :class:`~curses_fzf.ColorTheme` will be used.
            Default is ``None``.
        autoreturn (int): If :attr:`~curses_fzf.FuzzyFinder.autoreturn` is a positive integer,
            the :class:`~curses_fzf.FuzzyFinder` will automatically return the items
            in :attr:`~curses_fzf.FuzzyFinder.filtered` list without user input if
            the number of items in the filtered list matches the given number.
            If :attr:`~curses_fzf.FuzzyFinder.multi` is ``False``, the actual value
            of :attr:`~curses_fzf.FuzzyFinder.autoreturn` is ignored and the
            :class:`~curses_fzf.FuzzyFinder` will automatically return if there is
            exactly one item in the filtered list.
            Default is ``0``.
        page_size (int):  Number of items to move in :attr:`~curses_fzf.FuzzyFinder.filtered`
            list when pressing :kbd:`PAGE_UP`/:kbd:`PAGE_DOWN`.
            Default is ``10``.
        preview_window_percentage (int): :attr:`~curses_fzf.FuzzyFinder.preview_window_percentage`
            defines the width of the preview window as a percentage of the total width.
            Default is ``40``.
    """

    def __init__(self,
                 multi: bool = False,
                 title: str = "ITEMS",
                 query: str = "",
                 display: Callable[[Any], str] = lambda item: str(item),
                 preselect: Callable[[Any, ScoringResult], bool] = lambda item, result: False,
                 preview: Optional[Callable[[curses.window, ColorTheme, Any, ScoringResult], str]] = None,
                 score: Callable[[str, str], ScoringResult] = scoring_full_words,
                 color_theme: Optional[ColorTheme] = None,
                 autoreturn: int = 0,
                 page_size: int = 10,
                 preview_window_percentage: int = 40,
                 ) -> None:
        # user settings
        self.title: str = title
        """
        The :attr:`~curses_fzf.FuzzyFinder.title` to display in the upper left
        corner of the main :class:`~curses_fzf.FuzzyFinder` window.
        Default is ``"ITEMS"``.
        """
        self.autoreturn: int = autoreturn
        """
        If :attr:`~curses_fzf.FuzzyFinder.autoreturn` is a positive integer,
        the :class:`~curses_fzf.FuzzyFinder` will automatically return the items
        in :attr:`~curses_fzf.FuzzyFinder.filtered` list without user input if
        the number of items in the filtered list matches the given number.
        If :attr:`~curses_fzf.FuzzyFinder.multi` is ``False``, the actual value
        of :attr:`~curses_fzf.FuzzyFinder.autoreturn` is ignored and the
        :class:`~curses_fzf.FuzzyFinder` will automatically return if there is
        exactly one item in the filtered list.
        Default is ``0``.
        """
        self.preview_window_percentage: int = preview_window_percentage
        """
        :attr:`~curses_fzf.FuzzyFinder.preview_window_percentage` defines the
        width of the preview window as a percentage of the total width.
        The default value is ``40``.
        The preview window will be placed on the right side of the screen if a
        :meth:`~curses_fzf.FuzzyFinder.preview` function is provided.
        """
        self.page_size: int = page_size
        """
        Number of items to move in :attr:`~curses_fzf.FuzzyFinder.filtered` list
        when pressing :kbd:`PAGE_UP`/:kbd:`PAGE_DOWN`.
        Default is ``10``.
        """
        self.multi: bool = multi
        """
        :attr:`~curses_fzf.FuzzyFinder.multi` selection mode determines
        whether to allow selection of multiple items or not.
        Default is ``False``.
        """
        if color_theme is None:
            color_theme = ColorTheme()
        self.color_theme: ColorTheme = color_theme
        """
        The :attr:`~curses_fzf.FuzzyFinder.color_theme` to use for the interface.
        If ``None`` was given in the constructor, the default
        :class:`~curses_fzf.ColorTheme` will be used.
        """
        # function pointers
        self.display: Callable[[Any], str] = display
        """
        :meth:`~curses_fzf.FuzzyFinder.display` function is used to convert an
        item to a string for display and matching purposes.
        Default is ``lambda item: str(item)``.

        Args:
            item (Any): The item to convert to a string.

        Returns:
            str: The single-line string representation of the item.
        """
        self.preselect: Callable[[Any, ScoringResult], bool] = preselect
        """
        :meth:`~curses_fzf.FuzzyFinder.preselect` is a function to determine if
        an item should be preselected in :attr:`~curses_fzf.FuzzyFinder.multi`
        selection mode based on the item and its scoring result.
        Default is a function that always returns ``False``.

        Args:
            item (Any): The item from :attr:`~curses_fzf.FuzzyFinder.filtered`
                list to determine the preselection for.
            scoring_result (ScoringResult): The :class:`~curses_fzf.ScoringResult`
                of the item based on the current :attr:`~curses_fzf.FuzzyFinder.query`.

        Returns:
            bool: Whether the item should be preselected or not.
        """
        self.preview: Optional[Callable[[curses.window, ColorTheme, Any, ScoringResult], str]] = preview
        """
        If a :meth:`~curses_fzf.FuzzyFinder.preview` function is provided, a
        preview window will be shown for the currently highlighted item in the
        :attr:`~curses_fzf.FuzzyFinder.filtered` list.
        :attr:`~curses_fzf.FuzzyFinder.preview_window_percentage` defines the
        width of the preview window as a percentage of the total width.
        The preview window can be toggled with :kbd:`Ctrl + P`.

        The :meth:`~curses_fzf.FuzzyFinder.preview` function can be used in two ways:

        1. If the function returns a non-empty string, it will be rendered line
           by line inside the preview window, honoring the available space.
           In this case the :py:obj:`curses.window` parameter can be ignored.
        2. If the function returns an empty string, :class:`~curses_fzf.FuzzyFinder`
           will assume that the user is handling the rendering of the preview window
           using the provided :py:obj:`curses.window` parameter.
           In this case the user needs to take care of window boundaries.

        Args:
            preview_window (curses.window): The curses window to render the preview in.
            color_theme (ColorTheme): The :attr:`~curses_fzf.FuzzyFinder.color_theme`
                of :class:`~curses_fzf.FuzzyFinder`.
            item (Any): The item from :attr:`~curses_fzf.FuzzyFinder.filtered` list
                to generate the preview for.
            score_result (ScoringResult): The :class:`~curses_fzf.ScoringResult` of
                the item based on the current :attr:`~curses_fzf.FuzzyFinder.query`.

        Returns:
            str: The text to render in the preview window, if it is non-empty.
        """
        self.score: Callable[[str, str], ScoringResult] = score
        """
        The :meth:`~curses_fzf.FuzzyFinder.score` function is used to calculate
        the score of an item from :attr:`~curses_fzf.FuzzyFinder.all_items` list
        based on the current :attr:`~curses_fzf.FuzzyFinder.query` and the
        :meth:`~curses_fzf.FuzzyFinder.display` string of the item.
        The :attr:`~curses_fzf.ScoringResult.score` is used to sort the items in
        the :attr:`~curses_fzf.FuzzyFinder.filtered` list and to determine which
        items match the query.

        Default is :func:`~curses_fzf.scoring_full_words`.

        Args:
            query (str): The current :attr:`~curses_fzf.FuzzyFinder.query`.
                This is the string that the user has entered to filter the items.
            candidate (str): The string representation of the item as returned by
                :meth:`~curses_fzf.FuzzyFinder.display`.
                This is the string that is used for matching and display purposes.

        Returns:
            ScoringResult: The :class:`~curses_fzf.ScoringResult` of the item.
        """
        # internal state
        self.stdscr: Optional[curses.window] = None
        """
        The main curses window, will be set automatically in
        :class:`~curses_fzf.FuzzyFinder`'s main loop.
        """
        self.all_items: List[Any] = []
        """
        The original list of all items given by the user in
        :meth:`~curses_fzf.FuzzyFinder.find` method.
        This list will be filtered by :attr:`~curses_fzf.FuzzyFinder.query`
        into :attr:`~curses_fzf.FuzzyFinder.filtered` list.
        """
        self._preseed_query: str = query
        """
        Private: A saved copy of the initial query given on initialization,
        used to reset the query on :meth:`~curses_fzf.FuzzyFinder.find` call.
        """
        self._query: str = query
        """
        Private: Use the :attr:`~curses_fzf.FuzzyFinder.query` property to update the query.
        """
        self.show_preview: bool = True
        """
        Show or hide the preview window.
        This can be toggled with :kbd:`Ctrl + P` if a
        :meth:`~curses_fzf.FuzzyFinder.preview` function is provided.
        It will be set by :meth:`~curses_fzf.FuzzyFinder.kb_toggle_preview`.
        """
        self._cursor_items: int = 0
        """
        Private: Use the :attr:`~curses_fzf.FuzzyFinder.cursor_items` property
        to get the value.
        Use :attr:`~curses_fzf.FuzzyFinder.kb_move_items_cursor_absolute`
        and :attr:`~curses_fzf.FuzzyFinder.kb_move_items_cursor_relative` to
        update the cursor position.
        """
        self._cursor_query: int = len(self._query)
        """
        Private: Use the :attr:`~curses_fzf.FuzzyFinder.cursor_query` property
        to get the value.
        Use :attr:`~curses_fzf.FuzzyFinder.kb_move_query_cursor_absolute`
        and :attr:`~curses_fzf.FuzzyFinder.kb_move_query_cursor_relative` to
        update the cursor position.
        """
        self.return_selection_now: bool = False
        """
        Whether the :attr:`~curses_fzf.FuzzyFinder` should end the main loop
        and return the :attr:`~curses_fzf.FuzzyFinder.selected` items.
        This will be set to ``True`` by
        :meth:`~curses_fzf.FuzzyFinder.kb_accept_selection` on :kbd:`ENTER`.
        """
        self.filtered: List[Tuple[Any, ScoringResult]] = []
        """
        The list of items filtered by the current :attr:`~curses_fzf.FuzzyFinder.query`,
        each paired with its :class:`~curses_fzf.ScoringResult`.
        This list is updated by :meth:`~curses_fzf.FuzzyFinder.calculate_filtered`,
        which is called in each iteration of the main loop before rendering the items.
        """
        self.selected: List[Any] = []
        """
        The list of currently selected items in :attr:`~curses_fzf.FuzzyFinder.multi`
        selection mode.
        In single selection mode this list will be bypassed and the currently
        highlighted item in :attr:`~curses_fzf.FuzzyFinder.filtered` list will
        be returned on :kbd:`ENTER`.
        """
        # keymap
        self.keymap = {
            # ARROW-UP: move cursor up 1 position in filter list
            curses.KEY_UP: lambda: self.kb_move_items_cursor_relative(-1),  # 259
            # ARROW-DOWN: move cursor down 1 position in filter list
            curses.KEY_DOWN: lambda: self.kb_move_items_cursor_relative(1),  # 258
            # PAGE-UP: move cursor up by page_size positions
            curses.KEY_PPAGE: lambda: self.kb_move_items_cursor_relative(-self.page_size),  # 339
            # PAGE-DOWN: move cursor down by page_size positions
            curses.KEY_NPAGE: lambda: self.kb_move_items_cursor_relative(self.page_size),  # 338
            # HOME: move cursor to start of list
            curses.KEY_HOME: lambda: self.kb_move_items_cursor_absolute(0),  # 262
            # END: move cursor to end of list
            curses.KEY_END: lambda: self.kb_move_items_cursor_absolute(len(self.filtered) - 1),  # 360
            # ARROW-LEFT: move cursor left 1 position in query
            curses.KEY_LEFT: lambda: self.kb_move_query_cursor_relative(-1),  # 260
            # ARROW-RIGHT: move cursor right 1 position in query
            curses.KEY_RIGHT: lambda: self.kb_move_query_cursor_relative(1),  # 261
            # Ctrl + K: clear the query
            11: self.kb_reset_query,
            # BACKSPACE: remove the character before the cursor from the query
            8: self.kb_remove_from_query_cursor,  # ASCII backspace
            127: self.kb_remove_from_query_cursor,  # ASCII del
            curses.KEY_BACKSPACE: self.kb_remove_from_query_cursor,  # 263
            # DELETE: remove the character at the cursor from the query
            curses.KEY_DC: lambda: self.kb_remove_from_query_cursor(False),  # 330
            # ESC: raise abort exception
            27: self.kb_abort_selection,
            # ENTER: accept selection
            10: self.kb_accept_selection,  # linefeed (classic enter key)
            13: self.kb_accept_selection,  # carriage return (classic enter key)
            curses.KEY_ENTER: self.kb_accept_selection,  # 343
            # TAB: (de)select item (in multi mode)
            9: self.kb_toggle_selection,  # Tab - (de)select item (in multi mode)
            # Ctrl + A: select all from current filter (in multi mode)
            1: self.kb_select_all,
            # Ctrl + X: deselect all from current filter (in multi mode)
            24: self.kb_deselect_all,
            # Ctrl + P: toggle preview window
            16: self.kb_toggle_preview,
            # F1: show help
            curses.KEY_F1: self.kb_show_help,  # 265
        }
        """
        Dictionary mapping keys (e.g. ``curses.KEY_UP``) to their corresponding
        keybinding functions (e.g. ``lambda: self.kb_move_items_cursor_relative(-1)``).
        If the target function takes no arguments, it can be directly assigned
        like ``curses.KEY_F1: self.kb_show_help``.

        All the functions starting with ``kb_`` are keybinding functions that
        are used in the keymap and can also be reassigned or called directly.
        """

# properties

    @property
    def cursor_items(self) -> int:
        """
        The index of the cursor inside the :attr:`~curses_fzf.FuzzyFinder.filtered`
        list of items.
        Use :attr:`~curses_fzf.FuzzyFinder.kb_move_items_cursor_absolute`
        and :attr:`~curses_fzf.FuzzyFinder.kb_move_items_cursor_relative` to
        update the cursor position.
        """
        return self._cursor_items

    @property
    def cursor_query(self) -> int:
        """
        The index of the cursor inside the :attr:`~curses_fzf.FuzzyFinder.query`
        string.
        Use :attr:`~curses_fzf.FuzzyFinder.kb_move_query_cursor_absolute`
        and :attr:`~curses_fzf.FuzzyFinder.kb_move_query_cursor_relative` to
        update the cursor position.
        """
        return self._cursor_query

    @property
    def query(self) -> str:
        """
        The :attr:`~curses_fzf.FuzzyFinder.query` entered by the user or
        preseeded on initialization.
        Setting this property will also reset the
        :attr:`~curses_fzf.FuzzyFinder.cursor_items` cursor and move the
        :attr:`~curses_fzf.FuzzyFinder.cursor_query` cursor to the end of the query.
        """
        return self._query

    @query.setter
    def query(self, value: str) -> None:
        if self._query != value:
            self._cursor_items = 0
            self._cursor_query = len(value)
            self._query = value

# main entry point

    def find(self,
             items: List[Any],
             title: Optional[str] = None,
             query: Optional[str] = None,
             ) -> List[Any]:
        """
        Run the :class:`~curses_fzf.FuzzyFinder` on the given list of items
        and return the :attr:`~curses_fzf.FuzzyFinder.selected` item(s).

        Args:
            items (List[Any]): The list of items to filter and select from.
                This list will be stored in :attr:`~curses_fzf.FuzzyFinder.all_items`
                and filtered based on the :attr:`~curses_fzf.FuzzyFinder.query`
                into :attr:`~curses_fzf.FuzzyFinder.filtered` list.
            title (Optional[str]): The :attr:`~curses_fzf.FuzzyFinder.title` to
                display in the upper left corner of the main
                :class:`~curses_fzf.FuzzyFinder` window.
                Default is ``None``, in which case the title given on constructor
                will be reused.
            query (Optional[str]): The initial :attr:`~curses_fzf.FuzzyFinder.query`
                to preseed the interface with.
                Default is ``None``, in which case the query given on constructor
                will be reused.
                See :attr:`~curses_fzf.FuzzyFinder.query` for more details.

        Returns:
            List[Any]: The list of selected items.
        """
        self.all_items = items
        if title is not None:
            self.title = title
        if query is None:
            self._query = self._preseed_query
        else:
            self._query = query
        self._preseed_query = self._query
        # reset internal state
        self._cursor_items = 0
        self._cursor_query = len(self._query)
        self.show_preview = True
        self.return_selection_now = False
        self.filtered = []
        self.selected = []
        try:
            return curses.wrapper(lambda stdscr: self._main_loop(stdscr))
        except KeyboardInterrupt:
            raise CursesFzfAborted("fuzzyfinder aborted by user") from None

# keybinding functions

    def kb_move_items_cursor_absolute(self, position: int) -> None:
        """
        :attr:`~curses_fzf.FuzzyFinder.keymap` function:
        Move the :attr:`~curses_fzf.FuzzyFinder.cursor_items` cursor to the
        absolute :py:obj:`position` inside the :attr:`~curses_fzf.FuzzyFinder.filtered`
        list while keeping it within bounds of the list.

        Args:
            position (int): The absolute index inside the
                :attr:`~curses_fzf.FuzzyFinder.filtered` list to move the cursor to.
        """
        self._cursor_items = max(0, min(len(self.filtered) - 1, position))

    def kb_move_items_cursor_relative(self, offset: int) -> None:
        """
        :attr:`~curses_fzf.FuzzyFinder.keymap` function:
        Move the :attr:`~curses_fzf.FuzzyFinder.cursor_items` cursor by
        :py:obj:`offset` while keeping it within bounds of the
        :attr:`~curses_fzf.FuzzyFinder.filtered` list.

        Args:
            offset (int): The relative offset to move the cursor by inside the
                :attr:`~curses_fzf.FuzzyFinder.filtered` list.
                Negative values will move the cursor up, positive values will
                move it down.
        """
        self.kb_move_items_cursor_absolute(self.cursor_items + offset)

    def kb_move_query_cursor_absolute(self, position: int) -> None:
        """
        :attr:`~curses_fzf.FuzzyFinder.keymap` function:
        Move the :attr:`~curses_fzf.FuzzyFinder.cursor_query` cursor to the
        absolute :py:obj:`position` inside the :attr:`~curses_fzf.FuzzyFinder.query`
        string while keeping it within bounds of the string.

        Args:
            position (int): The absolute index inside the
                :attr:`~curses_fzf.FuzzyFinder.query` string to move the cursor to.
        """
        self._cursor_query = max(0, min(len(self.query), position))

    def kb_move_query_cursor_relative(self, offset: int) -> None:
        """
        :attr:`~curses_fzf.FuzzyFinder.keymap` function:
        Move the :attr:`~curses_fzf.FuzzyFinder.cursor_query` cursor by
        :py:obj:`offset` while keeping it within bounds of the
        :attr:`~curses_fzf.FuzzyFinder.query` string.

        Args:
            offset (int): The relative offset to move the cursor by inside the
                :attr:`~curses_fzf.FuzzyFinder.query` string.
                Negative values will move the cursor left, positive values will
                move it right.
        """
        self.kb_move_query_cursor_absolute(self.cursor_query + offset)

    def kb_abort_selection(self) -> None:
        """
        :attr:`~curses_fzf.FuzzyFinder.keymap` function:
        Abort the :class:`~curses_fzf.FuzzyFinder` and raise the
        :class:`~curses_fzf.CursesFzfAborted` exception.
        """
        raise CursesFzfAborted("fuzzyfinder aborted by user")

    def kb_accept_selection(self) -> None:
        """
        :attr:`~curses_fzf.FuzzyFinder.keymap` function:
        Accept the current :attr:`~curses_fzf.FuzzyFinder.selection` and return it.
        This will set :attr:`~curses_fzf.FuzzyFinder.return_selection_now` to ``True``.
        """
        # allow empty return in mutli mode but not in single mode
        if self.multi or self.filtered:
            self.return_selection_now = True

    def kb_toggle_preview(self) -> None:
        """
        :attr:`~curses_fzf.FuzzyFinder.keymap` function:
        Toggle the visibility of the :meth:`~curses_fzf.FuzzyFinder.show_preview`
        window.
        This will toggle the :attr:`~curses_fzf.FuzzyFinder.show_preview` boolean.
        """
        self.show_preview = not self.show_preview

    def kb_toggle_selection(self) -> None:
        """
        :attr:`~curses_fzf.FuzzyFinder.keymap` function:
        Toggle the selection state of the current item (only in
        :attr:`~curses_fzf.FuzzyFinder.multi` mode).
        This will add/remove the item to/from
        :attr:`~curses_fzf.FuzzyFinder.selected` list.
        """
        if self.multi and self.filtered:
            item = self.filtered[self.cursor_items][0]
            if item in self.selected:
                self.selected.remove(item)
            else:
                self.selected.append(item)

    def kb_select_all(self) -> None:
        """
        :attr:`~curses_fzf.FuzzyFinder.keymap` function:
        Select all items from the current :attr:`~curses_fzf.FuzzyFinder.filtered`
        list (only in :attr:`~curses_fzf.FuzzyFinder.multi` mode).
        """
        if self.multi:
            for entry in self.filtered:
                item = entry[0]
                if item not in self.selected:
                    self.selected.append(item)

    def kb_deselect_all(self) -> None:
        """
        :attr:`~curses_fzf.FuzzyFinder.keymap` function:
        Deselect all items from the current :attr:`~curses_fzf.FuzzyFinder.filtered`
        list (only in :attr:`~curses_fzf.FuzzyFinder.multi` mode).
        """
        if self.multi:
            for entry in self.filtered:
                item = entry[0]
                if item in self.selected:
                    self.selected.remove(item)

    def kb_reset_query(self) -> None:
        """
        :attr:`~curses_fzf.FuzzyFinder.keymap` function:
        Clear the :attr:`~curses_fzf.FuzzyFinder.query` string and return the
        :attr:`~curses_fzf.FuzzyFinder.cursor_query` and
        :attr:`~curses_fzf.FuzzyFinder.cursor_items` cursor to index ``0``.
        """
        if self.query:
            self._query = ""
            self._cursor_items = 0
            self._cursor_query = 0

    def kb_add_to_query(self, text: str, index: int = -1) -> None:
        """
        :attr:`~curses_fzf.FuzzyFinder.keymap` function:
        Add the given text to the :attr:`~curses_fzf.FuzzyFinder.query` before
        the given index.
        Adjust the :attr:`~curses_fzf.FuzzyFinder.cursor_query` cursor and reset
        the :attr:`~curses_fzf.FuzzyFinder.cursor_items` cursor if necessary.

        May raise :class:`~curses_fzf.CursesFzfIndexOutOfBounds` if the given
        index is out of bounds.

        Args:
            text (str): The text to add to the :attr:`~curses_fzf.FuzzyFinder.query`
                string.
            index (int): The index before which to add the text in the
                :attr:`~curses_fzf.FuzzyFinder.query` string.
                The default is ``-1``, which means to add the text at the end of
                the :attr:`~curses_fzf.FuzzyFinder.query` string.
                The index may also be negative, in which case it will be counted
                from the end of the string (e.g. ``-1`` means before the last
                character, ``-2`` means before the second last character, etc.).
        """
        if index < 0:
            index = len(self.query) + index + 1
        if not 0 <= index <= len(self.query):
            raise CursesFzfIndexOutOfBounds(
                "index to add to query is out of bounds")
        self._query = self.query[:index] + text + self.query[index:]
        # if the curser is before the insertion leave it where it is
        # otherwise move it according to insertion length
        if self.cursor_query >= index:
            self._cursor_query += len(text)
        # we will filter the list by typing a query,
        # so the old item cursor index is not valid anymore (it may leak out of
        # the list and even if it doesn't a completely other item may be selected)
        if text:
            self._cursor_items = 0

    def kb_add_to_query_cursor(self, text: str) -> None:
        """
        :attr:`~curses_fzf.FuzzyFinder.keymap` function:
        Add the given text to the query at the current
        :attr:`~curses_fzf.FuzzyFinder.cursor_query` cursor position.
        Adjust the :attr:`~curses_fzf.FuzzyFinder.cursor_query` cursor and
        reset the :attr:`~curses_fzf.FuzzyFinder.cursor_items` cursor if necessary.

        Args:
            text (str): The text to add to the :attr:`~curses_fzf.FuzzyFinder.query`
                string.
        """
        self.kb_add_to_query(text, self.cursor_query)

    def kb_remove_from_query(self, index: int = -1, length: int = 1) -> None:
        """
        :attr:`~curses_fzf.FuzzyFinder.keymap` function:
        Remove :py:obj:`length` characters from the query at position
        :py:obj:`index` and return the :attr:`~curses_fzf.FuzzyFinder.cursor_items`
        cursor to index ``0``.
        The :attr:`~curses_fzf.FuzzyFinder.cursor_query` cursor will be adjusted
        if it is after the remove index.

        Args:
            index (int): The index at which to remove characters from the
                :attr:`~curses_fzf.FuzzyFinder.query` string.
                The default is ``-1``, which means to remove the character before
                the end of the :attr:`~curses_fzf.FuzzyFinder.query` string.
                The index may also be negative, in which case it will be counted
                from the end of the string (e.g. ``-1`` means the last character,
                ``-2`` means the second last character, etc.).
            length (int): The number of characters to remove from the
                :attr:`~curses_fzf.FuzzyFinder.query` string starting from the
                given index. The default is ``1``. The length may be higher than the
                actual length of the query, but not negative.
        """
        if index < 0:
            index = len(self.query) + index
        if not 0 <= index < len(self.query):
            raise CursesFzfIndexOutOfBounds(
                "index to remove from query is out of bounds")
        if length < 0:
            raise CursesFzfIndexOutOfBounds(
                "length to remove from query my not be negative")
        if length > 0:
            remainder = self.query[index + length:]
            self._query = self.query[:index] + remainder
            self._cursor_items = 0
            # if the cursor is before or at the index leave it where it is
            if self.cursor_query > index:
                # if the cursor is in the removed part move it to the index
                if index <= self.cursor_query <= index + length:
                    self._cursor_query = index
                # otherwise move it according to removed length
                else:
                    self._cursor_query -= length

    def kb_remove_from_query_cursor(self, before: bool = True) -> None:
        """
        :attr:`~curses_fzf.FuzzyFinder.keymap` function:
        Remove one character from the :attr:`~curses_fzf.FuzzyFinder.query`
        before :attr:`~curses_fzf.FuzzyFinder.cursor_query` cursor position
        and return the :attr:`~curses_fzf.FuzzyFinder.cursor_items` cursor to
        index ``0``.
        The :attr:`~curses_fzf.FuzzyFinder.cursor_query` cursor will be adjusted.

        Args:
            before (bool): Whether to remove the character before the query cursor
                (like :kbd:`BACKSPACE`) or the character at the query cursor
                (like :kbd:`DEL`).
                Default is ``True``.
        """
        pos = self.cursor_query
        out_of_bounds = False
        if before:
            pos -= 1
            out_of_bounds = pos < 0
        else:
            out_of_bounds = pos >= len(self.query)
        if out_of_bounds:
            return
        self.kb_remove_from_query(pos, 1)

    def kb_show_help(self) -> None:
        """
        :attr:`~curses_fzf.FuzzyFinder.keymap` function:
        Show the help screen, using :kbd:`F1` key.
        """
        if self.stdscr:
            _help(self.stdscr, self.page_size, self.color_theme)

# loop functions

    def calculate_filtered(self) -> None:
        """
        Calculate the :attr:`~curses_fzf.FuzzyFinder.filtered` list of items
        from the :attr:`~curses_fzf.FuzzyFinder.all_items` list
        based on the current :attr:`~curses_fzf.FuzzyFinder.query`
        and :attr:`~curses_fzf.FuzzyFinder.score` function.

        This function will be called in each iteration of the main loop of
        :class:`~curses_fzf.FuzzyFinder` before rendering the items.
        """
        self.filtered = sorted(
            [
                (item, score_result) for item in self.all_items
                if (score_result := self.score(
                    self.query, self.display(item))) > 0
            ],
            key=lambda x: x[1], reverse=True
        )

    def _calculate_preselection(self) -> None:
        """
        Calculate the preselected items based on the current filter and
        preselection function.
        """
        if self.multi:
            self.selected = [item_tuple[0] for item_tuple in self.filtered if self.preselect(*item_tuple)]

    def _handle_input(self, key: UnicodeKey) -> None:
        """
        Handle the given key input by calling the corresponding keybinding function
        or adding the character to the query if it is a printable character.
        """
        int_key = key if isinstance(key, int) else ord(key)
        kb_function = self.keymap.get(int_key)
        if kb_function:
            kb_function()
        elif isinstance(key, str):
            if key.isprintable():
                self.kb_add_to_query_cursor(key)

    def _get_return_value(self) -> List[Any]:
        """
        Get the return value based on the current selection and multi mode.
        """
        # in multi mode return the list of selected items, even if it is empty
        # in single mode return the selected items list if not empty (e.g. if
        # the user manipulated it manually)
        if self.multi or self.selected:
            return self.selected
        # in single mode return the currently highlighted item if there is one,
        # otherwise an empty list
        return [self.filtered[self.cursor_items][0]] if self.filtered else []

    def _autoreturn(self) -> Optional[List[Any]]:
        """
        Check if the conditions for autoreturn are met and return the
        corresponding items.
        If is not 0 return directly if in single mode only 1 item is provided
        or left after initial filter.
        In multi mode the number of items need to match the number given
        as autoreturn's value.
        """
        if self.autoreturn:
            f_len = len(self.filtered)
            if self.multi:
                if f_len == self.autoreturn:
                    return [x[0] for x in self.filtered]
            elif f_len == 1:
                return [self.filtered[0][0]]
        return None

    def _render_query(self, width: int) -> None:
        """
        Render the query line based on the current query and query cursor.
        """
        if self.stdscr is None or width < 10:
            return
        # render query prompt
        self.stdscr.addstr(0, 2, "> ", curses.color_pair(self.color_theme.query))
        max_index = -1
        # render query characters with highlight on cursor position
        for i, c in enumerate(self.query[:width-6]):
            color = self.color_theme.query
            if self.cursor_query == i:
                color = self.color_theme.cursor
            self.stdscr.addstr(0, 4 + i, c, curses.color_pair(color))
            max_index = i
        # if the cursor is at the end of the query render a cursor symbol there
        if self.cursor_query == len(self.query) and self.cursor_query < width - 6:
            self.stdscr.addstr(0, 5 + max_index, " ", curses.color_pair(self.color_theme.cursor))

    def _render_no_match(self, width: int) -> None:
        """
        Render the "no match" message if there are no items matching the query.
        """
        if self.stdscr is None or width < 10:
            return
        if not self.filtered:
            self.stdscr.addstr(3, ITEM_COL_START + 2, "No matching items!"[:width-6],
                               curses.color_pair(self.color_theme.no_match))

    def _render_viewport(self, height: int, width: int) -> None:  # noqa: C901
        """
        Render the current viewport of the filtered items based on the current
        items cursor.
        """
        if self.stdscr is None or height < 7 or width < 10:
            return
        # query, footer, 2x lines and 1 empty line on top and bottom of list = 6
        viewport_height = height - 6
        viewport_start = max(0, self.cursor_items - viewport_height + 1)
        viewport_end = min(viewport_start + viewport_height, len(self.filtered))
        for i in range(viewport_start, viewport_end):
            # header, frame & empty line = 3
            row = i - viewport_start + 3
            item, score_result = self.filtered[i]
            display_item = self.display(item)
            if len(display_item.splitlines()) > 1:
                raise CursesFzfAssertion("display function must return single-line strings")
            # chose marker and color based on selection and cursor position
            marker = DESELECTED_MARKER
            base_color = self.color_theme.text
            if i == self.cursor_items and item in self.selected:
                marker = SELECTED_MARKER
                base_color = self.color_theme.cursor_selected
            elif i == self.cursor_items:
                base_color = self.color_theme.cursor
            elif item in self.selected:
                marker = SELECTED_MARKER
                base_color = self.color_theme.selected
            # render the marker before selected items
            self.stdscr.addstr(row, ITEM_COL_START, marker, curses.color_pair(base_color))
            # render the item character by character to highlight matched characters
            for char_index, char in enumerate(display_item[:width-10]):
                color = base_color
                for match in score_result.matches:
                    if match[0] <= char_index < match[0] + len(match[1]):
                        color = self.color_theme.highlight
                self.stdscr.addstr(row, ITEM_COL_START + 3 + char_index,
                                   char, curses.color_pair(color))
            # if the line is too long end it with "…"
            if len(display_item) > width - 10:
                self.stdscr.addstr(row, width - 6, CHAR_CONTINUED, curses.color_pair(base_color))

    def _render_preview(self, height: int, width: int) -> Optional[curses.window]:
        """
        Render the preview window if it is enabled and a preview function is provided.
        """
        if self.stdscr is None:
            return None
        # deactivate the preview window if the main window gets too small to display it properly
        if height < 7 or width < 30:
            self.show_preview = False
        sub_win = None
        if self.show_preview and self.preview is not None:
            sub_win = curses.newwin(
                height - 4,
                int(width * self.preview_window_percentage / 100),
                2,
                int(width * (100 - self.preview_window_percentage) / 100) - 2
            )
            sub_win.box()
            sub_win.addstr(0, 2, " PREVIEW ",
                           curses.color_pair(self.color_theme.window_title))
            if self.filtered:
                text = self.preview(sub_win, self.color_theme, self.filtered[self.cursor_items][0],
                                    self.filtered[self.cursor_items][1])
                # if the preview function returns any text assume the user didn't
                # use the preview_window parameter and render the text line by line
                # inside the preview window, honoring the available space
                if text:
                    sub_h, sub_w = sub_win.getmaxyx()
                    i = 2
                    for line in text.splitlines():
                        if i > sub_h - 3:
                            break
                        sub_win.addstr(i, 4, line[:sub_w - 6],
                                       curses.color_pair(self.color_theme.text))
                        i += 1
        return sub_win

    def _main_loop(self, stdscr: curses.window) -> List[Any]:
        self.stdscr = stdscr
        self.calculate_filtered()
        autoreturn_value = self._autoreturn()
        if autoreturn_value is not None:
            return autoreturn_value
        self._calculate_preselection()
        _init_curses()
        while True:
            self.calculate_filtered()
            # prepare window content
            height, width = _base_window(
                self.stdscr,
                self.title,
                (
                    f"{len(self.selected)} selected | "
                    f"{len(self.filtered)} matches | ↑↓ = navigate | "
                    f"{'TAB = toggle | ' if self.multi else ''}"
                    "ENTER = accept | ESC = abort | F1 = help"
                ),
                self.color_theme,
            )
            self._render_query(width)
            self._render_no_match(width)
            self._render_viewport(height, width)
            sub_win = self._render_preview(height, width)
            # render windows to screen
            self.stdscr.refresh()
            if sub_win is not None:
                sub_win.refresh()
            # read input
            self._handle_input(self.stdscr.get_wch())
            if self.return_selection_now:
                return self._get_return_value()
